import bpy
from bpy.props import (
    StringProperty,
    BoolProperty,
    BoolVectorProperty,
    IntProperty,
    IntVectorProperty,
    FloatProperty,
    FloatVectorProperty,
    EnumProperty,
    PointerProperty,
    CollectionProperty,
)


def get_volumetric_data(self, context):
    keys = bpy.data.collections[
        context.object.batoms.label
    ].batoms.settings_volumetric_data.keys()
    items = []
    i = 0
    for key in keys:
        items.append((key, key, key))
        i += 1
    return items


class Base(bpy.types.PropertyGroup):
    material_style: EnumProperty(
        name="material_style",
        description="Structural models",
        items=(
            ("default", "default", "default"),
            ("metallic", "metallic", "metallic"),
            ("plastic", "plastic", "plastic"),
            ("ceramic", "ceramic", "ceramic"),
            ("mirror", "mirror", "mirror"),
        ),
        default="default",
    )


class Belement(bpy.types.PropertyGroup):
    name: StringProperty(name="name", default="H")
    occupancy: FloatProperty(name="occupancy")
    radius: FloatProperty(name="radius", default=1)
    color: FloatVectorProperty(name="color", size=4, default=(0, 0.2, 0.8, 1))

    def as_dict(self) -> dict:
        setdict = {
            "occupancy": self.occupancy,
            "name": self.name,
            # 'radius': self.radius,
            "color": self.color[:],
        }
        return setdict


class Bspecies(Base):
    name: StringProperty(name="name", default="H")
    species: StringProperty(name="species", default="H")
    elements: CollectionProperty(name="Belements", type=Belement)
    scale: FloatProperty(name="scale", min=0, soft_max=2, default=1)
    radius: FloatProperty(name="radius", default=1.0)
    segments: IntVectorProperty(name="segments", size=2, default=(24, 16))
    color: FloatVectorProperty(
        name="color", size=4, subtype="COLOR", min=0, max=1, default=(0, 0.2, 0.8, 1)
    )

    @property
    def element_dict(self):
        element_dict = {}
        for ele in self.elements:
            element_dict[ele.name] = ele.as_dict()
        return element_dict

    def as_dict(self) -> dict:
        setdict = {
            "species": self.species,
            "name": self.name,
            "material_style": self.material_style,
            "color": self.color[:],
            "scale": self.scale,
            "elements": self.element_dict,
        }
        return setdict


class Batom(bpy.types.PropertyGroup):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="X")
    # select: StringProperty(name="select", default = 'all')
    species: CollectionProperty(name="Bspecies", type=Bspecies)
    radius: FloatProperty(name="radius")
    radii_style: EnumProperty(
        name="radii_style",
        description="Radii",
        items=(
            ("0", "Covalent", "covalent"),
            ("1", "VDW", "van der Waals"),
            ("2", "Ionic", "ionic"),
        ),
        default="0",
    )


class Battribute(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    name: StringProperty(name="name")
    label: StringProperty(name="label", default="batoms")
    data_type: EnumProperty(
        name="data_type",
        description="data type",
        items=(
            ("FLOAT", "FLOAT", "Floating-point value"),
            ("INT", "INT", "32-bit integer"),
            (
                "FLOAT_VECTOR",
                "FLOAT_VECTOR",
                "Vector – 3D vector with floating-point values.",
            ),
            (
                "FLOAT_COLOR",
                "FLOAT_COLOR",
                "Color – RGBA color with floating-point values.",
            ),
            ("BYTE_COLOR", "BYTE_COLOR", "Byte Color – RGBA color with 8-bit values."),
            ("STRING", "STRING", "String – Text string."),
            ("BOOLEAN", "BOOLEAN", "Boolean – True or false."),
            ("FLOAT2", "FLOAT2", "2D Vector – 2D vector with floating-point values."),
        ),
        default="FLOAT",
    )
    domain: EnumProperty(
        name="domain",
        description="Domain",
        items=(
            ("POINT", "POINT", ""),
            ("EDGE", "EDGE", ""),
            ("FACE", "FACE", ""),
        ),
        default="POINT",
    )
    dimension: IntProperty(name="index", default=1)
    shape_: IntVectorProperty(name="shape_", soft_min=0, size=32)
    delimiter: StringProperty(name="delimiter", default="@")

    @property
    def natt(self) -> int:
        import numpy as np

        return np.product(self.shape)

    @property
    def sub_name(self) -> int:
        sub_name = ["{}{}".format(self.name, i) for i in range(self.natt)]
        return sub_name

    @property
    def shape(self) -> int:
        return self.shape_[: self.dimension]

    @shape.setter
    def shape(self, data):
        tmp = list(self.shape_)
        tmp[0 : self.dimension] = list(data)
        self.shape_ = tmp

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "label": self.label,
            "name": self.name,
            "data_type": self.data_type,
            "domain": self.domain,
            "dimension": self.dimension,
            "shape": self.shape,
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "{:20s}{:10s}{:10s}{:10s}{:10s}   {:20s}\n".format(
            "Name", "Type", "Domain", "delimiter", "Dimension", "Shape"
        )
        s += "{:20s}{:10s}{:10s}{:10s}{:10d}  [".format(
            self.name, self.data_type, self.domain, self.delimiter, self.dimension
        )
        for i in range(self.dimension):
            s += "  {}  ".format(self.shape[i])
        s += "] \n"
        s += "-" * 80 + "\n"
        return s


class Bsite(bpy.types.PropertyGroup):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="X")
    index: IntProperty(name="index", default=0)


class Bcell(bpy.types.PropertyGroup):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="")
    pbc: BoolVectorProperty(name="pbc", default=[False, False, False], size=3)
    width: FloatProperty(name="width", default=0.05, min=0)
    color: FloatVectorProperty(
        name="color", size=4, subtype="COLOR", min=0, max=1, default=[0.2, 0.2, 0.2, 1]
    )

    def as_dict(self) -> dict:
        setdict = {
            "pbc": self.pbc,
            "width": self.width,
            "color": self.color[:],
        }
        return setdict


class Bvolumetric_data(bpy.types.PropertyGroup):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="")
    npoint: IntProperty(name="npoint")
    shape: IntVectorProperty(name="shape", size=3)

    def as_dict(self) -> dict:
        setdict = {
            "label": self.label,
            "shape": self.shape[:],
        }
        return setdict


class Bboundary(Base):
    """ """

    name: StringProperty(name="name")
    flag: BoolProperty(name="flag", default=False)
    active: BoolProperty(name="active", default=False)
    label: StringProperty(name="label", default="batoms")
    boundary: FloatVectorProperty(
        name="boundary", default=[0.0, 1.0, 0.0, 1.0, 0.0, 1.0], size=6
    )
    scale: FloatProperty(name="scale", soft_min=0.01, soft_max=1, default=0.01)
    color: FloatVectorProperty(
        name="color", size=4, subtype="COLOR", min=0, max=1, default=[0, 1, 1, 0.5]
    )

    def as_dict(self) -> dict:
        setdict = {
            "name": self.name,
            "flag": self.flag,
            "label": self.label,
            "active": self.active,
            "color": self.color[:],
            "boundary": self.boundary[:],
            "scale": self.scale,
            "material_style": self.material_style,
        }
        return setdict

    def __repr__(self) -> str:
        import numpy as np

        s = "-" * 60 + "\n"
        s = "Name    scale   color  \n"
        s += "{:6s}  {:1.3f}  [{:1.2f}  {:1.2f}  {:1.2f}   {:1.2f}] \n".format(
            self.name,
            self.scale,
            self.color[0],
            self.color[1],
            self.color[2],
            self.color[3],
        )
        boundary = np.array(self.boundary)
        s += " {} \n ".format(boundary)
        s += "-" * 60 + "\n"
        return s


class Bselect(bpy.types.PropertyGroup):
    """ """

    flag: BoolProperty(name="flag", default=False)
    name: StringProperty(name="name", default="all")
    model_style: EnumProperty(
        name="model_style",
        description="Structural models",
        items=(
            ("0", "Space-filling", "Use ball"),
            ("1", "Ball-and-stick", "Use ball and stick"),
            ("2", "Polyhedral", "Use polyhedral"),
            ("3", "Wireframe", "Use wireframe"),
        ),
        default="0",
    )
    radius_style: EnumProperty(
        name="radius_style",
        description="Radii",
        items=(
            ("0", "Covalent", "covalent"),
            ("1", "VDW", "van der Waals"),
            ("2", "Ionic", "ionic"),
        ),
        default="0",
    )
    polyhedra_style: EnumProperty(
        name="polyhedra_style",
        description="Polhhedra models",
        items=(
            ("0", "atoms, bonds and polyhedra", "atoms, bonds and polyhedra"),
            ("1", "atoms, polyhedra", "atoms, polyhedra"),
            ("2", "central atoms, polyhedra", "central atoms, polyhedra"),
            ("3", "polyhedra", "polyhedra"),
        ),
        default="0",
    )
    bspecies: CollectionProperty(name="Bspecies", type=Bspecies)

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name   model_style   radius_style     show \n"
        s += "{0:10s} {1:10s}   {2:10s}   {3:10s} \n".format(
            self.name, str(self.model_style), self.radius_style, str(self.show)
        )
        s += "-" * 60 + "\n"
        return s


items = [
    ("OTHER", "Not Batoms", "Not Batoms"),
    ("BATOMS", "Batoms", "Batoms"),
    ("VOLUME", "Volume", "Volume"),
    ("CELL", "Cell", "Cell"),
    ("BOND", "Bond", "Bond"),
    ("BOUNDARY", "Boundary", "Boundary"),
    ("INSTANCER", "Instancer", "Instancer"),
    ("POLYHEDRA", "Polyhedra", "Polyhedra"),
    ("CRYSTALSHAPE", "Crystal Shape", "Crystal Shape"),
    ("LATTICEPLANE", "LatticePlane", "LatticePlane"),
    ("ISOSURFACE", "Isosurface", "Isosurface"),
    ("MS", "MS", "Molecular Surface"),
    ("CAVITY", "CAVITY", "Cavity"),
    ("MARGES", "Marges", "NMR tensors"),
    ("HIGHLIGHT", "HIGHLIGHT", "Highlight"),
]


class BatomsCollection(bpy.types.PropertyGroup):
    """ """

    label: StringProperty(name="label", default="batoms")
    type: EnumProperty(
        name="type", description="Structural models", items=items, default="OTHER"
    )

    model_style: EnumProperty(
        name="model_style",
        description="Structural models",
        items=(
            ("0", "Space-filling", "Use ball"),
            ("1", "Ball-and-stick", "Use ball and stick"),
            ("2", "Polyhedral", "Use polyhedral"),
            ("3", "Wireframe", "Use wireframe"),
        ),
        default="0",
    )

    radius_style: EnumProperty(
        name="radius_style",
        description="Radii",
        items=(
            ("0", "Covalent", "covalent"),
            ("1", "VDW", "van der Waals"),
            ("2", "Ionic", "ionic"),
        ),
        default="0",
    )

    color_style: EnumProperty(
        name="color_style",
        description="Color",
        items=(("0", "JMOL", "JMOL"), ("1", "VESTA", "VESTA"), ("2", "CPK", "CPK")),
        default="0",
    )

    polyhedra_style: EnumProperty(
        name="polyhedra_style",
        description="Polhhedra models",
        items=(
            ("0", "atoms, bonds and polyhedra", "atoms, bonds and polyhedra"),
            ("1", "atoms, polyhedra", "atoms, polyhedra"),
            ("2", "central atoms, polyhedra", "central atoms, polyhedra"),
            ("3", "polyhedra", "polyhedra"),
        ),
        default="0",
    )

    segments: IntVectorProperty(name="segments", size=2, default=(24, 16))
    scale: FloatProperty(name="scale", default=1)
    show: BoolProperty(name="show", default=True)
    show_unit_cell: BoolProperty(name="show_unit_cell", default=True)
    show_axes: BoolProperty(name="show_axes", default=True)
    show_label: StringProperty(name="show_label", default="")
    wrap: BoolVectorProperty(name="wrap", default=[False, False, False], size=3)
    boundary: PointerProperty(name="Bboundary", type=Bboundary)
    cell: PointerProperty(name="Bcell", type=Bcell)
    crystal_view: BoolProperty(name="crystal_view", default=False)
    ui_list_index_species: IntProperty(name="ui_list_index_species", default=0)
    ui_list_index_select: IntProperty(name="ui_list_index_select", default=0)
    ui_list_index_volumetric_data: IntProperty(
        name="ui_list_index_volumetric_data", default=0
    )
    # collection
    settings_select: CollectionProperty(name="settings_select", type=Bselect)

    settings_species: CollectionProperty(name="settings_species", type=Bspecies)
    settings_volumetric_data: CollectionProperty(
        name="settings_volumetric_data", type=Bvolumetric_data
    )

    def as_dict(self) -> dict:
        setdict = {
            "model_style": self.model_style,
            "color_style": self.color_style,
            "radius_style": self.radius_style,
            "polyhedra_style": self.polyhedra_style,
            "segments": self.segments,
            "wrap": self.wrap,
            "show_axes": self.show_axes,
            "show_label": self.show_label,
            "crystal_view": self.crystal_view,
            "scale": self.scale,
        }
        return setdict


class BatomsObject(bpy.types.PropertyGroup):
    label: StringProperty(name="label", default="batoms")
    type: EnumProperty(
        name="type", description="Structural models", items=items, default="OTHER"
    )
    # obj
    atom: PointerProperty(name="Batom", type=Batom)
    cell: PointerProperty(name="Bcell", type=Bcell)
    volume: PointerProperty(name="Bvolumetric_data", type=Bvolumetric_data)

    # collection
    settings_attribute: CollectionProperty(name="settings_attribute", type=Battribute)



"""

Lattice Planes

To insert lattice planes in structural models.
"""
import bpy
from batoms.base.collection import Setting, tuple2string
import logging

logger = logging.getLogger(__name__)


class LatticePlaneSettings(Setting):
    """
    PlaneSetting object

    The PlaneSetting object store the polyhedra information.

    Parameters:

    label: str
        The label define the batoms object that a Setting belong to.

    """

    def __init__(self, label, parent=None, plane=None) -> None:
        Setting.__init__(self, label, coll_name="%s" % label)
        self.name = "Blatticeplane"
        self.parent = parent
        if plane is not None:
            for key, data in plane.items():
                self[key] = data

    def get_collection(self):
        if self.coll_name:
            coll = bpy.data.collections.get(self.coll_name)
            collection = getattr(coll, self.name)
        elif self.obj_name:
            obj = bpy.data.objects.get(self.obj_name)
            collection = getattr(obj, self.name)
        else:
            raise KeyError("The collection property {} not exist!".format(self.name))
        return collection.settings

    @property
    def no(
        self,
    ):
        return self.parent.batoms.get_spacegroup_number()

    @no.setter
    def no(self, no):
        self.no = no

    def __setitem__(self, key, setdict):
        """
        Set properties
        """
        subset = self.find(key)
        if subset is None:
            subset = self.bpy_setting.add()
            name = tuple2string(key)
            subset.name = name
            self.ui_list_index = len(self) - 1
        subset.indices = key
        subset.name = name
        subset.flag = True
        for key, value in setdict.items():
            setattr(subset, key, value)
        subset.label = self.label

    def add(self, indices):
        self[indices] = {"indices": indices}

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Indices   distance  slicing   boundary\n"
        for p in self.bpy_setting:
            s += "{0:10s}   {1:1.3f}   ".format(p.name, p.distance)
            s += "{:10s}  {:10s}  \n".format(str(p.slicing), str(p.boundary))
        s += "-" * 60 + "\n"
        return s



"""
"""
import bpy
from batoms.base.collection import Setting
import logging

logger = logging.getLogger(__name__)


class TemplateSettings(Setting):
    def __init__(self, label, parent) -> None:
        """Template object
        This object store the plugin setting.

        Parameters:

        label: str
            The label define the batoms object that
            a Setting belong to.
        """
        Setting.__init__(self, label, coll_name=label)
        self.label = label
        self.name = "Btemplate"
        self.parent = parent

    def add(self, name, datas={}):
        self[name] = datas

    def set_collection(self, label):
        """ """
        if not bpy.data.collections.get(label):
            coll = bpy.data.collections.new(label)
            self.parent.batoms.coll.children.link(coll)
            coll.batoms.type = "PTemplate"
            coll.batoms.label = label

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "name  type select\n"
        for setting in self.bpy_setting:
            s += "{:4s}   {:1.2f}".format(setting.name, setting.prop1)
        s += "-" * 60 + "\n"
        return s



import bpy
from batoms.batoms import Batoms
import numpy as np


def test_settings(ch4):
    """species panel"""
    assert ch4.coll.batoms.ui_list_index_select == 0
    # add
    ch4.selects.add("s1", [1])
    assert ch4.coll.batoms.ui_list_index_select == 1
    # remove
    ch4.selects.remove("s1")
    assert ch4.coll.batoms.ui_list_index_select == 0


def test_select():
    from ase.build import molecule, fcc111
    from batoms.batoms import Batoms
    import numpy as np

    bpy.ops.batoms.delete()
    au111 = fcc111("Au", (4, 4, 4), vacuum=0)
    au111 = Batoms("au111", from_ase=au111)
    mol = Batoms("mol", from_ase=molecule("CH3CH2OH"))
    mol.translate([5, 5, 10])
    au111 = au111 + mol
    au111.cell[2, 2] += 10
    assert len(au111.selects) == 3
    # model_style
    au111.selects["mol"].model_style = 1
    assert len(au111.bond.bondlists) == 8
    assert np.isclose(au111[-1].scale, 0.4)
    au111.selects["mol"].model_style = 0
    assert np.isclose(au111[-1].scale, 1)
    assert len(au111.bond.bondlists) == 0


def test_select_protein():
    from batoms.pdbparser import read_pdb

    bpy.ops.batoms.delete()
    atoms = read_pdb("../tests/datas/1ema.pdb")  # 1tim
    batoms = Batoms("protein", from_ase=atoms)
    batoms.ribbon.draw()
    sel1 = batoms.selects.add("sel1", np.where(batoms.arrays["types"] == "HETATM")[0])
    sel1.show = 1
    sel1.model_style = 1
    bpy.ops.batoms.delete()



"""Test if batoms' default preferences and startup file are correctly loaded
This test assumes the user has initialized the Blender environment with default settings
"""
import bpy
import numpy as np


def test_pref_load():
    """Test several preference settings that are significantly
    from factory Blender
    """
    bpy.ops.batoms.use_batoms_startup()
    bpy.ops.batoms.use_batoms_preference()
    # Default gradiant for theme
    DEFAULT_GRAD = 0.18823
    DEFAULT_HI_GRAD = 0.2392
    CUSTOM_GRAD = 0.50
    CUSTOM_HI_GRAD = 0.90
    theme = bpy.context.preferences.themes[0]
    gradient = theme.view_3d.space.gradients.gradient[0]
    hi_gradient = theme.view_3d.space.gradients.high_gradient[0]
    # Checking if a lighter theme is set
    assert not np.isclose(hi_gradient, DEFAULT_HI_GRAD, atol=0.08)
    assert not np.isclose(gradient, DEFAULT_GRAD, atol=0.08)
    assert np.isclose(hi_gradient, CUSTOM_HI_GRAD, atol=0.08)
    assert np.isclose(gradient, CUSTOM_GRAD, atol=0.08)
    # Batoms-related
    assert "batoms" in bpy.context.preferences.addons
    addon = bpy.context.preferences.addons["batoms"]
    assert addon.preferences.logging_level == "WARNING"
    return


def test_startup_load():
    # Check Layout
    # Check Scene unit system
    bpy.ops.batoms.use_batoms_startup()
    bpy.ops.batoms.use_batoms_preference()
    scene = bpy.context.scene
    assert scene.unit_settings.system == "NONE"
    assert scene.unit_settings.scale_length == 1.0
    # Check no default objects exists
    assert len(bpy.data.objects) == 0
    assert len(bpy.data.collections) == 1
    return



from .lattice_plane import LatticePlane

from . import (
    bpy_data,
    ops,
    gui,
)

__all__ = ["LatticePlane"]

classes_bpy_data = [
    # internal data first
    bpy_data.LatticePlaneSetting,
    bpy_data.LatticePlane,
    gui.LatticePlaneProperties,
]
classes = [
    ops.LatticePlaneAdd,
    ops.LatticePlaneRemove,
    ops.LatticePlaneDraw,
    ops.LatticePlaneModify,
    gui.VIEW3D_PT_Batoms_lattice_plane,
    gui.BATOMS_MT_lattice_plane_context_menu,
    gui.BATOMS_UL_lattice_plane,
    gui.BATOMS_PT_lattice_plane,
]


def register_class():
    from bpy.types import Collection, Object, Scene
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    #
    for cls in classes_bpy_data:
        register_class(cls)
    # attach to blender internal data
    Collection.Blatticeplane = PointerProperty(
        name="Blatticeplane", type=bpy_data.LatticePlane
    )
    Object.Blatticeplane = PointerProperty(
        name="Blatticeplane", type=bpy_data.LatticePlane
    )
    Scene.Blatticeplane = PointerProperty(type=gui.LatticePlaneProperties)
    #
    for cls in classes:
        register_class(cls)


def unregister_class():
    from bpy.types import Collection, Object, Scene
    from bpy.utils import unregister_class

    del Collection.Blatticeplane
    del Object.Blatticeplane
    del Scene.Blatticeplane
    for cls in reversed(classes):
        unregister_class(cls)
    for cls in reversed(classes_bpy_data):
        unregister_class(cls)



import bpy
from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    EnumProperty,
    FloatVectorProperty,
    CollectionProperty,
)

from batoms.internal_data import Base


class IsosurfaceSetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    name: StringProperty(name="name")
    label: StringProperty(name="label", default="")
    level: FloatProperty(name="level", default=0.10)
    color: FloatVectorProperty(
        name="color",
        size=4,
        subtype="COLOR",
        min=0,
        max=1,
        default=[1, 1, 0, 0.5],
    )
    color1: FloatVectorProperty(
        name="color1", size=4, subtype="COLOR", min=0, max=1, default=[1, 0, 0, 1.0]
    )
    color2: FloatVectorProperty(
        name="color2", size=4, subtype="COLOR", min=0, max=1, default=[0, 0, 1, 1.0]
    )

    def get_volumetric_data(self, context):
        keys = bpy.data.collections[self.label].batoms.settings_volumetric_data.keys()
        # items = [("0", "None", "None")]
        items = []
        i = 1
        for key in keys:
            items.append((key, key, key))
            i += 1
        return items

    volumetric_data: EnumProperty(
        items=get_volumetric_data,
        name="volumetric_data",
        description="choose a volumetric_data",
        default=None,
        update=None,
    )

    def get_color_by(self, context):
        keys = bpy.data.collections[self.label].batoms.settings_volumetric_data.keys()
        items = [("None", "None", "None")]
        i = 1
        for key in keys:
            items.append((key, key, key))
            i += 1
        return items

    color_by: EnumProperty(
        items=get_color_by,
        name="color_by",
        description="color by",
        default=None,
        update=None,
    )

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "label": self.label,
            "name": self.name,
            "volumetric_data": self.volumetric_data,
            "color_by": self.color_by,
            "material_style": self.material_style,
            "color": self.color[:],
            "level": self.level,
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name    volumetric_data    level        color            \n"
        s += "{0:10s}   {1:10s}  {2:1.6f}  [{3:1.2f}  {4:1.2f}  {5:1.2f}   {6:1.2f}] \n".format(
            self.name,
            self.volumetric_data,
            self.level,
            self.color[0],
            self.color[1],
            self.color[2],
            self.color[3],
        )
        s += "-" * 60 + "\n"
        return s


class Isosurface(bpy.types.PropertyGroup):
    """This module defines the Isosurface properties to extend
    Blender’s internal data.

    """

    active: BoolProperty(name="active", default=False)
    settings: CollectionProperty(name="IsosurfaceSetting", type=IsosurfaceSetting)

    ui_list_index: IntProperty(name="ui_list_index", default=0)

    def as_dict(self) -> dict:
        setdict = {}
        return setdict



import bpy
from bpy.types import (
    Panel,
    Operator,
)
from bpy.props import (
    BoolProperty,
    IntVectorProperty,
    FloatProperty,
    FloatVectorProperty,
)
from batoms.utils.butils import get_selected_batoms, get_selected_objects
from batoms.batoms import Batoms


# The panel.
class Plane_PT_prepare(Panel):
    bl_label = "Plane"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Plugins"
    bl_idname = "PLANE_PT_Tools"

    def draw(self, context):
        plane = context.scene.batoms.plane

        layout = self.layout
        row = layout.row()
        row.prop(plane, "indices")
        layout.prop(plane, "distance")
        layout.prop(plane, "scale")
        layout.prop(plane, "symmetry", expand=True)
        row = layout.row()
        row.prop(plane, "crystal")
        row.prop(plane, "center")
        layout.prop(plane, "slicing", expand=True)
        layout.prop(plane, "boundary", expand=True)
        layout.prop(plane, "show_edge")
        layout.prop(plane, "color")
        layout.operator("batoms.add_plane")


class PlaneProperties(bpy.types.PropertyGroup):
    @property
    def selected_batoms(self):
        return get_selected_batoms()

    @property
    def selected_plane(self):
        return get_selected_objects("bplane")

    def Callback_modify_indices(self, context):
        plane = bpy.context.scene.batoms.plane
        modify_plane_attr(
            self.selected_batoms, self.selected_plane, "indices", plane.indices
        )

    def Callback_modify_distance(self, context):
        plane = bpy.context.scene.batoms.plane
        modify_plane_attr(
            self.selected_batoms, self.selected_plane, "distance", plane.distance
        )

    def Callback_modify_scale(self, context):
        plane = bpy.context.scene.batoms.plane
        modify_plane_attr(
            self.selected_batoms, self.selected_plane, "scale", plane.scale
        )

    def Callback_modify_crystal(self, context):
        plane = bpy.context.scene.batoms.plane
        modify_plane_attr(
            self.selected_batoms,
            self.selected_plane,
            "crystal",
            plane.crystal,
            center=plane.center,
        )

    def Callback_modify_center(self, context):
        plane = bpy.context.scene.batoms.plane
        modify_plane_attr(
            self.selected_batoms, self.selected_plane, "center", plane.center
        )

    def Callback_modify_symmetry(self, context):
        plane = bpy.context.scene.batoms.plane
        modify_plane_attr(
            self.selected_batoms, self.selected_plane, "symmetry", plane.symmetry
        )

    def Callback_modify_slicing(self, context):
        plane = bpy.context.scene.batoms.plane
        modify_plane_attr(
            self.selected_batoms, self.selected_plane, "slicing", plane.slicing
        )

    def Callback_modify_boundary(self, context):
        plane = bpy.context.scene.batoms.plane
        modify_plane_attr(
            self.selected_batoms, self.selected_plane, "boundary", plane.boundary
        )

    def Callback_modify_show_edge(self, context):
        plane = bpy.context.scene.batoms.plane
        show_edge = plane.show_edge
        modify_plane_attr(
            self.selected_batoms, self.selected_plane, "show_edge", show_edge
        )

    def Callback_modify_color(self, context):
        plane = bpy.context.scene.batoms.plane
        color = plane.color
        modify_plane_attr(self.selected_batoms, self.selected_plane, "color", color)

    indices: IntVectorProperty(
        name="Miller indices",
        size=3,
        default=(0, 0, 1),
        description="Miller indices for the plane",
        update=Callback_modify_indices,
    )
    distance: FloatProperty(
        name="Distance",
        default=3,
        description="distance from origin",
        update=Callback_modify_distance,
    )
    scale: FloatProperty(
        name="Scale",
        default=1,
        description="scale of the plane",
        update=Callback_modify_scale,
    )
    crystal: BoolProperty(
        name="crystal",
        default=False,
        description="plane to form crystal shape",
        update=Callback_modify_crystal,
    )
    center: BoolProperty(
        name="center",
        default=False,
        description="Apply center to model",
        update=Callback_modify_center,
    )
    symmetry: BoolProperty(
        name="symmetry",
        default=False,
        description="Apply symmetry to indices",
        update=Callback_modify_symmetry,
    )
    slicing: BoolProperty(
        name="slicing",
        default=False,
        description="Apply slicing to volumetric data",
        update=Callback_modify_slicing,
    )
    boundary: BoolProperty(
        name="boundary",
        default=False,
        description="Apply boundary to model",
        update=Callback_modify_boundary,
    )
    show_edge: BoolProperty(
        name="show_edge",
        default=False,
        description="show_edge",
        update=Callback_modify_show_edge,
    )
    color: FloatVectorProperty(
        name="color",
        subtype="COLOR",
        default=(0.1, 0.8, 0.4, 0.8),
        size=4,
        description="color picker",
        update=Callback_modify_color,
    )


def modify_plane_attr(batoms_name_list, plane_name_list, key, value, center=False):
    selected_plane_new = []
    for batoms_name in batoms_name_list:
        batoms = Batoms(label=batoms_name)
        for plane_name in plane_name_list:
            plane = bpy.data.objects[plane_name]
            if plane.bplane.label == batoms_name:
                batoms.planesetting[plane.bplane.indices] = {key: value}
            selected_plane_new.append(plane_name)
        batoms.planesetting.draw_lattice_plane()
        if center:
            batoms.planesetting.draw_crystal_shape(origin=batoms.cell.center)
        else:
            batoms.planesetting.draw_crystal_shape()
    for name in selected_plane_new:
        obj = bpy.data.objects.get(name)
        if obj is not None:
            obj.select_set(True)


def add_plane(
    indices,
    color,
    distance,
    scale,
    crystal,
    symmetry,
    slicing,
    boundary,
    show_edge,
    center=False,
):
    """ """
    selected_batoms = get_selected_batoms()
    for batoms_name in selected_batoms:
        batoms = Batoms(label=batoms_name)
        batoms.planesetting[indices] = {
            "indices": indices,
            "distance": distance,
            "scale": scale,
            "crystal": crystal,
            "symmetry": symmetry,
            "slicing": slicing,
            "boundary": boundary,
            "show_edge": show_edge,
            "color": color,
        }
        batoms.planesetting.draw_lattice_plane()
        if center:
            batoms.planesetting.draw_crystal_shape(origin=batoms.cell.center)
        else:
            batoms.planesetting.draw_crystal_shape()


class AddButton(Operator):
    bl_idname = "batoms.add_plane"
    bl_label = "Add"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add plane"

    indices: IntVectorProperty(
        name="Miller indices",
        size=3,
        default=(0, 0, 1),
        description="Miller indices for the plane",
    )
    distance: FloatProperty(
        name="Distance",
        default=3,
        description="distance from origin",
    )
    scale: FloatProperty(
        name="Scale",
        default=1,
        description="scale of the plane",
    )
    crystal: BoolProperty(
        name="crystal",
        default=False,
        description="plane to form crystal shape",
    )
    center: BoolProperty(
        name="center",
        default=False,
        description="Apply center to model",
    )
    symmetry: BoolProperty(
        name="symmetry",
        default=False,
        description="Apply symmetry to indices",
    )
    slicing: BoolProperty(
        name="slicing",
        default=False,
        description="Apply slicing to volumetric data",
    )
    boundary: BoolProperty(
        name="boundary",
        default=False,
        description="Apply boundary to model",
    )
    show_edge: BoolProperty(
        name="show_edge",
        default=False,
        description="show_edge",
    )
    color: FloatVectorProperty(
        name="color",
        subtype="COLOR",
        default=(0.1, 0.8, 0.4, 0.8),
        size=4,
        description="color picker",
    )

    def execute(self, context):
        add_plane(
            self.indices,
            self.color,
            self.distance,
            self.scale,
            self.crystal,
            self.symmetry,
            self.slicing,
            self.boundary,
            self.show_edge,
            self.center,
        )
        return {"FINISHED"}



import bpy
import numpy as np

try:
    from _common_helpers import has_display

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_settings():
    """key search"""
    from batoms.batoms import Batoms
    from ase.io.cube import read_cube_data

    bpy.ops.batoms.delete()
    volume, atoms = read_cube_data("../tests/datas/h2o-homo.cube")
    h2o = Batoms("h2o", from_ase=atoms, volume={"homo": volume})
    assert h2o.volumetric_data.bpy_setting["homo"].shape[:] == volume.shape
    assert h2o.volumetric_data["homo"].shape == volume.shape
    assert len(h2o.volumetric_data) == 1
    h2o.volumetric_data["electrostatic"] = volume + 1
    assert len(h2o.volumetric_data) == 2
    assert h2o.volumetric_data.find("homo") is not None
    assert h2o.volumetric_data.find("electrostatic") is not None
    h2o.volumetric_data.remove("homo")
    assert h2o.volumetric_data.find("homo") is None


def test_gui():
    """gui panel"""
    from batoms.batoms import Batoms
    from ase.io.cube import read_cube_data

    bpy.ops.batoms.delete()
    data, atoms = read_cube_data("../tests/datas/h2o-homo.cube")
    h2o = Batoms("h2o", from_ase=atoms)
    assert h2o.coll.batoms.ui_list_index_volumetric_data == 0
    h2o.volumetric_data["homo"] = data
    assert h2o.coll.batoms.ui_list_index_volumetric_data == 0
    #
    hartree, atoms = read_cube_data("../tests/datas/h2o-hartree.cube")
    h2o.volumetric_data["hartree"] = -hartree
    assert h2o.coll.batoms.ui_list_index_volumetric_data == 1
    #
    h2o.volumetric_data["diff"] = (
        h2o.volumetric_data["hartree"] - h2o.volumetric_data["homo"]
    )
    assert h2o.coll.batoms.ui_list_index_volumetric_data == 2


def test_ops_add():
    """gui panel"""
    from batoms.batoms import Batoms
    from ase.io.cube import read_cube_data

    bpy.ops.batoms.delete()
    data, atoms = read_cube_data("../tests/datas/h2o-homo.cube")
    h2o = Batoms("h2o", from_ase=atoms)
    assert len(h2o.volumetric_data) == 0
    bpy.context.view_layer.objects.active = h2o.obj
    bpy.ops.batoms.volumetric_data_add(
        name="hartree", filepath="../tests/datas/h2o-hartree.cube"
    )
    assert h2o.volumetric_data.find("hartree") is not None
    bpy.ops.batoms.volumetric_data_remove(name="hartree")
    assert h2o.volumetric_data.find("hartree") is None


def test_ops_create(h2o_homo):
    from ase.io.cube import read_cube_data

    h2o = h2o_homo
    bpy.context.view_layer.objects.active = h2o.obj
    hartree, atoms = read_cube_data("../tests/datas/h2o-hartree.cube")
    h2o.volumetric_data["hartree"] = -hartree
    bpy.ops.batoms.volumetric_data_create(
        name="diff", select_data1="h2o_homo", select_data2="h2o_homo", operator="Minus"
    )
    assert h2o.coll.batoms.ui_list_index_volumetric_data == 2
    assert np.isclose(h2o.volumetric_data["diff"], 0).all()


def test_repeat():
    from batoms.bio.bio import read

    bpy.ops.batoms.delete()
    h2o = read("../tests/datas/h2o-homo.cube")
    shape = h2o.volumetric_data["h2o_homo"].shape
    M = [2, 2, 1]
    h2o._volumetric_data *= M
    new_shape = h2o.volumetric_data["h2o_homo"].shape
    assert list(new_shape) == [shape[i] * M[i] for i in range(3)]



from .crystal_shape import CrystalShape

from . import (
    bpy_data,
    ops,
    gui,
)

__all__ = ["CrystalShape"]

classes_bpy_data = [
    # internal data first
    bpy_data.CrystalShapeSetting,
    bpy_data.CrystalShape,
    gui.CrystalShapeProperties,
]
classes = [
    ops.CrystalShapeAdd,
    ops.CrystalShapeRemove,
    ops.CrystalShapeDraw,
    ops.CrystalShapeModify,
    gui.VIEW3D_PT_Batoms_crystal_shape,
    gui.BATOMS_MT_crystal_shape_context_menu,
    gui.BATOMS_UL_crystal_shape,
    gui.BATOMS_PT_crystal_shape,
]


def register_class():
    from bpy.types import Collection, Object, Scene
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    #
    for cls in classes_bpy_data:
        register_class(cls)
    # attach to blender internal data
    Collection.Bcrystalshape = PointerProperty(
        name="Bcrystalshape", type=bpy_data.CrystalShape
    )
    Object.Bcrystalshape = PointerProperty(
        name="Bcrystalshape", type=bpy_data.CrystalShape
    )
    Scene.Bcrystalshape = PointerProperty(type=gui.CrystalShapeProperties)
    #
    for cls in classes:
        register_class(cls)


def unregister_class():
    from bpy.types import Collection, Object, Scene
    from bpy.utils import unregister_class

    del Collection.Bcrystalshape
    del Object.Bcrystalshape
    del Scene.Bcrystalshape
    for cls in reversed(classes):
        unregister_class(cls)
    for cls in reversed(classes_bpy_data):
        unregister_class(cls)



import bpy
from batoms.cell import Bcell
import numpy as np


def test_cell(au):
    """ """
    cell = au.cell
    assert isinstance(cell, Bcell)


def test_cell_transform(au):
    au.cell = [2, 2, 2]
    au.cell[2, 2] += 5
    assert au.cell[2, 2] == 7
    au.cell.translate([0, 0, 2])
    assert au.cell[2, 2] == 7
    assert au.cell.global_positions[3, 2] == 9
    # au.cell.rotate(np.pi/4, 'Z')
    # assert au.cell[2, 2] == 7
    # assert au.cell.positions[3, 2] == 9


def test_repeat(au):
    au.cell = [2, 2, 2]
    # au.cell.rotate(np.pi/4, 'Z')
    au.cell.repeat([2, 2, 2])
    assert np.allclose(au.cell.array, np.array([[4, 0, 0], [0, 4, 0], [0, 0, 4]]))


def test_translate_repeat(au):
    au.cell = [2, 2, 2]
    au.cell.translate([0, 0, 2])
    au.cell.repeat([2, 2, 2])
    assert np.allclose(au.cell.array, np.array([[4, 0, 0], [0, 4, 0], [0, 0, 4]]))


def test_copy(au):
    au.cell = [2, 2, 2]
    cell2 = au.cell.copy("au-2")
    assert isinstance(cell2, Bcell)
    reciprocal = au.cell.reciprocal
    assert np.isclose(reciprocal[0, 0], 3.14159)


def test_draw(au):
    au.cell = [2, 2, 2]
    au.cell.width = 0.1
    au.cell.color = [1, 0, 0, 1]


def test_bond_reload(au):
    """save to blend file and reload"""
    import os
    from batoms import Batoms

    au.cell.width = 0.01
    au.cell.color = [1, 0, 0, 0.5]
    cwd = os.getcwd()
    filepath = os.path.join(cwd, "test.blend")
    bpy.ops.wm.save_as_mainfile(filepath=filepath)
    bpy.ops.batoms.delete()
    bpy.ops.wm.open_mainfile(filepath=filepath)
    au = Batoms("au")
    assert np.isclose(au.cell.width, 0.01)
    assert np.isclose(au.cell.color, np.array([1, 0, 0, 0.5])).all()



"""
"""

import bpy
from bpy.types import Menu, Panel, UIList
from bpy.props import (
    StringProperty,
    EnumProperty,
)

from batoms.internal_data.bpy_data import get_volumetric_data


class VolumetricDataProperties(bpy.types.PropertyGroup):
    # volumetric data
    name: StringProperty(
        name="name",
        description="Name of volumetric data",
        default="create",
    )
    select_data1: EnumProperty(
        name="select_data1",
        description="Select volumetric data",
        items=get_volumetric_data,
        default=0,
    )
    select_data2: EnumProperty(
        name="select_data2",
        description="Select volumetric data",
        items=get_volumetric_data,
        default=0,
    )
    operator: EnumProperty(
        name="operator",
        items=[("Add", "Add", "", 0), ("Minus", "Minus", "", 1)],
        description="Volumetric data to be used.",
        default=0,
    )


class VIEW3D_PT_Batoms_volumetric_data(Panel):
    bl_label = "Volumetric Data"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Batoms"
    bl_idname = "VIEW3D_PT_Batoms_volumetric_data"
    bl_options = {"DEFAULT_CLOSED"}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type != "OTHER"
        else:
            return False

    def draw(self, context):
        name = "None"
        if context.object:
            if context.object.batoms.type != "OTHER":
                name = context.object.batoms.label
        layout = self.layout
        layout.label(text="Active: " + name)
        batoms = context.scene.batoms
        op = layout.operator(
            "batoms.volumetric_data_create",
            icon="GREASEPENCIL",
            text="Create from others",
        )
        op.name = batoms.volumetric_data.name
        op.select_data1 = batoms.volumetric_data.select_data1
        op.select_data2 = batoms.volumetric_data.select_data2
        op.operator = batoms.volumetric_data.operator
        layout.separator()


class BATOMS_MT_volumetric_data_context_menu(Menu):
    bl_label = "Volumetric Data Specials"
    bl_idname = "BATOMS_MT_volumetric_data_context_menu"

    def draw(self, _context):
        layout = self.layout
        op = layout.operator("batoms.volumetric_data_add", icon="ADD", text="Add Data")
        layout.separator()
        op = layout.operator(
            "batoms.volumetric_data_remove", icon="X", text="Delete All Data"
        )
        op.all = True


class BATOMS_UL_volumetric_data(UIList):
    def draw_item(
        self, _context, layout, _data, item, icon, active_data, _active_propname, index
    ):
        volumetric_data = item
        custom_icon = "OBJECT_DATAMODE"
        if self.layout_type in {"DEFAULT", "COMPACT"}:
            split = layout.split(factor=0.66, align=False)
            split.prop(volumetric_data, "name", text="", emboss=False, icon=custom_icon)
            row = split.row(align=True)
            row.emboss = "NONE_OR_STATUS"
            # row.prop(volumetric_data, "distance", text="")
        elif self.layout_type == "GRID":
            layout.alignment = "CENTER"
            layout.label(text="", icon=custom_icon)


class BATOMS_PT_volumetric_data(Panel):
    bl_label = "Volumetric_data"
    bl_category = "Batoms"
    bl_idname = "BATOMS_PT_volumetric_data"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_parent_id = "VIEW3D_PT_Batoms_volumetric_data"
    # bl_options = {'DEFAULT_CLOSED'}

    COMPAT_ENGINES = {"BLENDER_RENDER", "BLENDER_EEVEE", "BLENDER_WORKBENCH"}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type != "OTHER"
        else:
            return False

    def draw(self, context):
        layout = self.layout

        ob = context.object
        ba = bpy.data.collections[ob.batoms.label].batoms
        if len(ba.settings_volumetric_data) > 0:
            kb = ba.settings_volumetric_data[ba.ui_list_index_volumetric_data]
        else:
            kb = None

        row = layout.row()

        rows = 3
        if kb:
            rows = 5

        row.template_list(
            "BATOMS_UL_volumetric_data",
            "",
            ba,
            "settings_volumetric_data",
            ba,
            "ui_list_index_volumetric_data",
            rows=rows,
        )

        col = row.column(align=True)
        op = col.operator("batoms.volumetric_data_add", icon="ADD", text="")
        op = col.operator("batoms.volumetric_data_remove", icon="REMOVE", text="")
        if kb is not None:
            op.name = kb.name
        col.separator()

        col.menu(
            "BATOMS_MT_volumetric_data_context_menu", icon="DOWNARROW_HLT", text=""
        )

        if kb:
            col.separator()

            sub = col.column(align=True)

            split = layout.split(factor=0.4)
            row = split.row()

            row = split.row()
            row.alignment = "RIGHT"

            sub = row.row(align=True)
            sub.label()  # XXX, for alignment only

            sub = row.row()
            layout.use_property_split = True
            row = layout.row()
            col = layout.column()
            sub = col.column(align=True)
            # sub.prop(kb, "radius_style", text="Radius_style")
            col.prop(kb, "shape", text="Shape")



"""
"""
import bpy
import numpy as np
from ase.cell import Cell
from batoms.utils.butils import object_mode
from batoms.base.object import ObjectGN


class Bcell(ObjectGN):
    def __init__(
        self,
        label,
        array=None,
        batoms=None,
        location=np.array([0, 0, 0]),
    ) -> None:
        """
        Unit cell of three dimensions.
        ver: 3x3 verlike object
          The three cell vectors: cell[0], cell[1], and cell[2].
        """
        self.label = label
        name = "cell"
        self.batoms = batoms
        ObjectGN.__init__(self, label, name)
        if array is not None:
            self.build_object(array, location)
        # if self.show_axes:
        # self.show_axes = True

    def build_object(self, array, location):
        """
        Draw unit cell by edge, however, can not be rendered.
        """
        array = np.array(array)
        positions = np.zeros([4, 3])
        if array is None:
            self._trajectory = {"positions": []}
        elif array.shape in [(3,), (3, 1), (6,), (6, 1), (3, 3)]:
            cell = Cell.new(array)
            positions[1:4, :] = cell.array
            self._trajectory = {"positions": [positions]}
        elif len(array.shape) == 3:
            positions = array[0]
            self._trajectory = {"positions": [positions]}
        else:
            raise Exception("Shape of cell %s is not corrected!" % array.shape)

        self.delete_obj(self.obj_name)
        mesh = bpy.data.meshes.new(self.obj_name)
        mesh.from_pydata(positions, [[0, 1], [0, 2], [0, 3]], [])
        mesh.update()
        for f in mesh.polygons:
            f.use_smooth = True
        obj = bpy.data.objects.new(self.obj_name, mesh)
        obj.data = mesh
        obj.location = location
        obj.batoms.type = "CELL"
        obj.batoms.label = self.label
        if self.batoms is not None:
            self.batoms.coll.objects.link(obj)
        else:
            bpy.data.collections["Collection"].objects.link(obj)
        # materials
        mat = self.build_materials(self.label, color=self.color)
        obj.data.materials.append(mat)
        self.init_geometry_node_modifier()
        self.build_geometry_node()
        self.set_attributes({"positions": positions})
        bpy.context.view_layer.update()

    def build_geometry_node(self):
        """ """
        from batoms.utils.butils import get_node_by_name

        links = self.gn_node_group.links
        nodes = self.gn_node_group.nodes
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        # ------------------------------------------------------------------
        # transfer first 4 positions of cell
        PositionCell = get_node_by_name(
            nodes, "%s_PositionCell" % (self.label), "GeometryNodeInputPosition"
        )
        TransferCells = []
        for i in range(4):
            TransferCell = get_node_by_name(
                nodes, "%s_TransferCell_%s" % (self.label, i), "GeometryNodeSampleIndex"
            )
            TransferCell.data_type = "FLOAT_VECTOR"
            InputInt = get_node_by_name(
                nodes, "%s_InputInt_%s" % (self.label, i), "FunctionNodeInputInt"
            )
            InputInt.integer = i
            links.new(GroupInput.outputs["Geometry"], TransferCell.inputs[0])
            links.new(PositionCell.outputs["Position"], TransferCell.inputs[3])
            links.new(InputInt.outputs[0], TransferCell.inputs["Index"])
            TransferCells.append(TransferCell)
        # ------------------------------------------------------------------
        VectorAdds = []
        for i in range(4):
            VectorAdd = get_node_by_name(
                nodes, "%s_VectorAdd_%s" % (self.label, i), "ShaderNodeVectorMath"
            )
            VectorAdd.operation = "ADD"
            VectorAdds.append(VectorAdd)
        links.new(TransferCells[1].outputs[2], VectorAdds[0].inputs[0])
        links.new(TransferCells[2].outputs[2], VectorAdds[0].inputs[1])
        links.new(TransferCells[1].outputs[2], VectorAdds[1].inputs[0])
        links.new(TransferCells[3].outputs[2], VectorAdds[1].inputs[1])
        links.new(TransferCells[2].outputs[2], VectorAdds[2].inputs[0])
        links.new(TransferCells[3].outputs[2], VectorAdds[2].inputs[1])
        links.new(TransferCells[3].outputs[2], VectorAdds[3].inputs[0])
        links.new(VectorAdds[0].outputs[0], VectorAdds[3].inputs[1])
        # ============================================================
        # In this implementation.
        # Use an object with four vertices to represent a cell, instead of eight vertices.
        # We use curve to mesh methods to draw the edges (cylinder)
        # One does not need to call cell.draw() anymore.
        # circle for profle of curve
        Circle = get_node_by_name(
            nodes, "%s_Circle" % self.label, "GeometryNodeCurvePrimitiveCircle"
        )
        Circle.inputs[4].default_value = self.width
        # faces from vertices
        faces = [
            [0, 2, 6, 3],
            [1, 4, 7, 5],
            [0, 1, 4, 2],
            [3, 5, 7, 6],
            [0, 1, 5, 3],
            [2, 4, 7, 6],
        ]
        input_nodes = TransferCells + VectorAdds
        for i in range(6):
            # set quadrilaterial
            # maybe in the future, we have curve primitive: cube
            Quadrilateral = get_node_by_name(
                nodes,
                "%s_Quadrilateral_%s" % (self.label, i),
                "GeometryNodeCurvePrimitiveQuadrilateral",
            )
            Quadrilateral.mode = "POINTS"
            CurveToMesh = get_node_by_name(
                nodes, "%s_CurveToMesh_%s" % (self.label, i), "GeometryNodeCurveToMesh"
            )
            links.new(Quadrilateral.outputs[0], CurveToMesh.inputs[0])
            links.new(Circle.outputs[0], CurveToMesh.inputs[1])
            links.new(CurveToMesh.outputs[0], JoinGeometry.inputs[0])
            for j in range(4):
                k = 2 if faces[i][j] < 4 else 0
                links.new(
                    input_nodes[faces[i][j]].outputs[k], Quadrilateral.inputs[j + 7]
                )
        setMaterial = get_node_by_name(
            self.gn_node_group.nodes,
            "%s_setMaterial" % (self.label),
            "GeometryNodeSetMaterial",
        )
        setMaterial.inputs[2].default_value = self.material
        links.new(JoinGeometry.outputs[0], setMaterial.inputs[0])
        links.new(setMaterial.outputs[0], GroupOutput.inputs[0])
        # The order of join geometry seems to be inverse.
        # So we output the original geometry in the last step.
        # Thus, the first four vertices are the vertices of the cell.
        links.new(GroupInput.outputs["Geometry"], JoinGeometry.inputs["Geometry"])

    @property
    def material(self):
        return self.get_material()

    def get_material(self):
        name = "%s_cell" % self.label
        mat = bpy.data.materials.get(name)
        return mat

    def __repr__(self) -> str:
        numbers = np.round(self.array, 3).tolist()
        s = "Cell({})".format(numbers)
        return s

    def __getitem__(self, index):
        return self.local_array[index]

    def __setitem__(self, index, value):
        """Set unit cell vectors.

        Parameters:

        Examples:

        """
        positions = self.positions
        positions[1:4, :][index] = value
        self.positions = positions
        self.batoms.update_gn_cell()
        self.batoms.boundary.update_gn_cell()

    def __array__(self, dtype=float):
        if dtype != float:
            raise ValueError("Cannot convert cell to array of type {}".format(dtype))
        return self.local_array

    @property
    def width(self):
        return self.batoms.coll.batoms.cell.width

    @width.setter
    def width(self, width):
        from batoms.utils.butils import get_node_by_name

        self.batoms.coll.batoms.cell.width = width
        Circle = get_node_by_name(
            self.gn_node_group.nodes,
            "%s_Circle" % (self.label),
            "GeometryNodeCurvePrimitiveCircle",
        )
        Circle.inputs[4].default_value = width

    @property
    def color(self):
        # Viewpoint_color = self.materials[self.main_element].diffuse_color
        # for node in self.material.node_tree.nodes:
        #     if 'Base Color' in node.inputs:
        #         node_color = node.inputs['Base Color'].default_value[:]
        #     if 'Alpha' in node.inputs:
        #         Alpha = node.inputs['Alpha'].default_value
        # color = [node_color[0], node_color[1], node_color[2], Alpha]
        color = self.batoms.coll.batoms.cell.color
        return color

    @color.setter
    def color(self, color):
        from batoms.utils.butils import get_node_by_name

        if len(color) == 3:
            color = [color[0], color[1], color[2], 1]
        self.batoms.coll.batoms.cell.color = color
        self.material.diffuse_color = color
        for node in self.material.node_tree.nodes:
            if "Base Color" in node.inputs:
                node.inputs["Base Color"].default_value = color
            if "Alpha" in node.inputs:
                node.inputs["Alpha"].default_value = color[3]
        setMaterial = get_node_by_name(
            self.gn_node_group.nodes,
            "%s_setMaterial" % (self.label),
            "GeometryNodeSetMaterial",
        )
        setMaterial.inputs[2].default_value = self.material

    def get_trajectory(self, local=True):
        """ """
        trajectory = {"positions": self.get_shape_key(self.obj, local=local)}
        return trajectory

    def set_trajectory(self, trajectory=None, frame_start=0):
        if trajectory is None:
            trajectory = self._trajectory
        nframe = len(trajectory["positions"])
        if nframe == 0:
            return
        name = self.label
        obj = self.obj
        self.set_shape_key(name, obj, trajectory["positions"], frame_start=frame_start)

    @property
    def local_array(self):
        return self.get_local_array()

    def get_local_array(self):
        """
        In this case, the origin is translated to vertices[0].
        While the orientation is reserved.

        Returns:
            (3x3 local_array): The array of cell.
        """
        positions = self.positions
        local_array = np.array(
            [
                positions[1] - positions[0],
                positions[2] - positions[0],
                positions[3] - positions[0],
            ]
        )
        return local_array

    @property
    def array(self):
        return self.get_array()

    def get_array(self):
        """
        In this case, the origin is translated to vertices[0].
        While the orientation is reserved.

        Returns:
            (3x3 array): The array of cell.
        """
        positions = self.positions
        array = np.array(
            [
                positions[1] - positions[0],
                positions[2] - positions[0],
                positions[3] - positions[0],
            ]
        )
        return array

    @property
    def origin(self):
        return self.positions[0]

    @property
    def edges(self):
        """Edges of the cell"""
        edge_indices = [
            [0, 3],
            [0, 1],
            [4, 2],
            [4, 1],
            [3, 5],
            [2, 6],
            [7, 5],
            [7, 6],
            [0, 2],
            [3, 6],
            [1, 5],
            [4, 7],
        ]
        basis = np.array(
            [
                [0, 0, 0],
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [1, 1, 0],
                [1, 0, 1],
                [0, 1, 1],
                [1, 1, 1],
            ]
        )
        positions = np.dot(basis, self.array)
        edges = []
        for indices in edge_indices:
            edges.append(positions[indices])
        return edges

    def copy(self, label):
        object_mode()
        cell = Bcell(
            label, batoms=self.batoms, array=self.array, location=self.obj.location
        )
        return cell

    def repeat(self, m):
        self[:] = np.array([m[c] * self.local_array[c] for c in range(3)])

    @property
    def length(self):
        length = np.linalg.norm(self.local_array, axis=1)
        return length

    @property
    def reciprocal(self):
        from math import pi

        b1 = 2 * pi / self.volume * np.cross(self[1], self[2])
        b2 = 2 * pi / self.volume * np.cross(self[2], self[0])
        b3 = 2 * pi / self.volume * np.cross(self[0], self[1])
        return np.array([b1, b2, b3])

    @property
    def volume(self):
        return np.dot(self[0], np.cross(self[1], self[2]))

    @property
    def center(self):
        """Center of unit cell."""
        return (self.array[0] + self.array[1] + self.array[2]) / 2.0

    def build_materials(
        self,
        label,
        color=None,
        node_type="Principled BSDF",
        use_smooth=True,
        node_inputs=None,
        material_style="plastic",
        backface_culling=True,
    ):
        """ """
        from batoms.material import create_material

        name = "%s_cell" % (label)
        if color is None:
            color = [0.2, 0.2, 0.2, 1]
        self.delete_material(name)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=True,
        )
        return mat

    @property
    def draw_crystal_axes(self):
        """draw_crystal_axes object.
        Shoud be a global variable.
        """
        from batoms.draw.draw_screen import DrawCrystalAxes

        dns = bpy.app.driver_namespace
        name = "{}_crystal_axes".format(self.label)
        if self.label in dns:
            return dns[self.label]
        context = bpy.context
        dns[name] = DrawCrystalAxes(context, name)
        return dns[name]

    @property
    def show_axes(self):
        if self.batoms is not None:
            return self.batoms.coll.batoms.show_axes
        else:
            return False

    @show_axes.setter
    def show_axes(self, show_axes):
        self.batoms.coll.batoms.show_axes = show_axes
        positions = self.array
        self.draw_crystal_axes.remove_handle()
        if show_axes:
            self.draw_crystal_axes.add_handle(positions)

    def draw(self):
        from batoms.utils import deprecated

        deprecated(
            '"draw" will be deprecated in the furture. The cell is drawn automaticely now.'
        )
        pass

    def as_dict(self):
        """ """
        data = {
            "array": self.array,
        }
        data.update(self.batoms.coll.batoms.cell.as_dict())
        return data



from .template import Template

from . import (
    bpy_data,
    ops,
    gui,
)

__all__ = ["Template"]

classes_bpy_data = [
    bpy_data.TemplateSetting,
    bpy_data.Template,
    gui.TemplateProperties,
]
classes = [
    ops.TemplateAdd,
    ops.TemplateRemove,
    ops.TemplateDraw,
    gui.VIEW3D_PT_Batoms_template,
    gui.BATOMS_MT_template_context_menu,
    gui.BATOMS_UL_template,
    gui.BATOMS_PT_template,
]


def register_class():
    from bpy.types import Collection, Object, Scene
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    for cls in classes_bpy_data:
        register_class(cls)
    # attach to blender internal data
    Collection.Btemplate = PointerProperty(name="Btemplate", type=bpy_data.Template)
    Object.Btemplate = PointerProperty(name="Btemplate", type=bpy_data.Template)
    Scene.Btemplate = PointerProperty(type=gui.TemplateProperties)
    #
    for cls in classes:
        register_class(cls)


def unregister_class():
    from bpy.types import Collection, Object, Scene
    from bpy.utils import unregister_class

    #
    del Collection.Btemplate
    del Object.Btemplate
    del Scene.Btemplate
    for cls in reversed(classes):
        unregister_class(cls)
    for cls in reversed(classes_bpy_data):
        unregister_class(cls)



import bpy
from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    FloatVectorProperty,
    CollectionProperty,
)

from batoms.internal_data import Base


class SheetSetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="")
    name: StringProperty(name="name")
    # sheetId: IntProperty(name="sheetId", default=0)
    # chainId: StringProperty(name="chainId", default='A')
    startChain: StringProperty(name="startChain")
    startResi: IntProperty(name="startResi", default=0)
    endChain: StringProperty(name="endChain")
    endResi: IntProperty(name="endResi", default=0)
    color: FloatVectorProperty(name="color1", size=4, default=(0.0, 0.0, 1.0, 1.0))
    extrude: FloatProperty(name="extrude", default=0.8)
    depth: FloatProperty(name="depth", default=0.1)

    @property
    def name(self) -> str:
        return "%s-%s-%s-%s" % (
            self.startChain,
            self.startResi,
            self.endChain,
            self.endResi,
        )

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "label": self.label,
            "startChain": self.startChain,
            "startResi": self.startResi,
            "endChain": self.endChain,
            "endResi": self.endResi,
            "name": self.name,
            "color": self.color[:],
            "extrude": self.extrude,
            "depth": self.depth,
            "material_style": self.material_style,
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name   startChain   startResi   endChain   endResi\n"
        s += "{0:10s} {1:10s}   {2:10s}      {3:10s}   {4:10s} \n".format(
            self.name,
            self.startChain,
            str(self.startResi),
            self.endChain,
            str(self.endResi),
        )
        s += "-" * 60 + "\n"
        return s


class HelixSetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="")
    name: StringProperty(name="name")
    # helixId: IntProperty(name="helixId", default=0)
    # chainId: StringProperty(name="chainId", default='A')
    startChain: StringProperty(name="startChain")
    startResi: IntProperty(name="startResi", default=0)
    endChain: StringProperty(name="endChain")
    endResi: IntProperty(name="endResi", default=0)
    color: FloatVectorProperty(name="color1", size=4, default=(1.0, 0.0, 0.0, 1))
    extrude: FloatProperty(name="extrude", default=1.0)
    depth: FloatProperty(name="depth", default=0.1)

    @property
    def name(self) -> str:
        return "%s-%s-%s-%s" % (
            self.startChain,
            self.startResi,
            self.endChain,
            self.endResi,
        )

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "label": self.label,
            "startChain": self.startChain,
            "startResi": self.startResi,
            "endChain": self.endChain,
            "endResi": self.endResi,
            "name": self.name,
            "color": self.color[:],
            "extrude": self.extrude,
            "depth": self.depth,
            "material_style": self.material_style,
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name   startChain   startResi   endChain   endResi\n"
        s += "{0:10s} {1:10s}   {2:10s}      {3:10s}   {4:10s} \n".format(
            self.name,
            self.startChain,
            str(self.startResi),
            self.endChain,
            str(self.endResi),
        )
        s += "-" * 60 + "\n"
        return s


class TurnSetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="")
    name: StringProperty(name="name")
    # turnId: IntProperty(name="turnId", default=0)
    # chainId: StringProperty(name="chainId", default='A')
    startChain: StringProperty(name="startChain")
    startResi: IntProperty(name="startResi", default=0)
    endChain: StringProperty(name="endChain")
    endResi: IntProperty(name="endResi", default=0)
    color: FloatVectorProperty(name="color1", size=4, default=(0.0, 1.0, 0.0, 1))
    radius: FloatProperty(name="radius", default=0.2)

    @property
    def name(self) -> str:
        return "%s-%s-%s-%s" % (
            self.startChain,
            self.startResi,
            self.endChain,
            self.endResi,
        )

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "label": self.label,
            "startChain": self.startChain,
            "startResi": self.startResi,
            "endChain": self.endChain,
            "endResi": self.endResi,
            "name": self.name,
            "color": self.color[:],
            "radius": self.radius,
            "material_style": self.material_style,
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name   startChain   startResi   endChain   endResi\n"
        s += "{0:10s} {1:10s}   {2:10s}      {3:10s}   {4:10s} \n".format(
            self.name,
            self.startChain,
            str(self.startResi),
            self.endChain,
            str(self.endResi),
        )
        s += "-" * 60 + "\n"
        return s


class Protein(bpy.types.PropertyGroup):
    """This module defines the Protein properties to extend
    Blender’s internal data.

    """

    turn_ui_list_index: IntProperty(name="turn_ui_list_index", default=0)
    sheet_ui_list_index: IntProperty(name="sheet_ui_list_index", default=0)
    helix_ui_list_index: IntProperty(name="helix_ui_list_index", default=0)

    # colleciton properties
    settings_turn: CollectionProperty(name="settings_turn", type=TurnSetting)
    settings_sheet: CollectionProperty(name="settings_sheet", type=SheetSetting)
    settings_helix: CollectionProperty(name="settings_helix", type=HelixSetting)



import bpy
from bpy.types import Panel
from bpy.props import (
    BoolProperty,
    FloatProperty,
    FloatVectorProperty,
    IntVectorProperty,
)

from batoms import Batoms
from batoms.gui.utils import (
    get_attr,
    set_attr,
    get_active_bpy_data_batoms,
    set_module_attr,
)


class Cell_PT_prepare(Panel):
    bl_label = "Cell"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Batoms"
    bl_idname = "BATOMS_PT_Cell"

    def draw(self, context):
        layout = self.layout
        cell = context.scene.batoms.cell

        box = layout.box()
        box.label(text="Basis")
        row = box.row(align=True)
        row.prop(cell, "cell_a0")
        row.prop(cell, "cell_a1")
        row.prop(cell, "cell_a2")
        row = box.row(align=True)
        row.prop(cell, "cell_b0")
        row.prop(cell, "cell_b1")
        row.prop(cell, "cell_b2")
        row = box.row(align=True)
        row.prop(cell, "cell_c0")
        row.prop(cell, "cell_c1")
        row.prop(cell, "cell_c2")
        layout.prop(cell, "width")
        layout.prop(cell, "color", text="color")
        layout.prop(cell, "pbc")

        op = layout.operator("batoms.apply_transform")
        op.a = cell.transform_a
        op.b = cell.transform_b
        op.c = cell.transform_c
        layout.prop(cell, "transform_a")
        layout.prop(cell, "transform_b")
        layout.prop(cell, "transform_c")
        #
        op = layout.operator("batoms.apply_boundary")
        op.a = cell.boundary_a
        op.b = cell.boundary_b
        op.c = cell.boundary_c
        layout.prop(cell, "boundary_a")
        layout.prop(cell, "boundary_b")
        layout.prop(cell, "boundary_c")


def get_cell(i1, i2):
    """Helper function to easily get cell-property."""

    def getter(self):
        context = bpy.context
        if context.object and context.object.batoms.type != "OTHER":
            obj = bpy.data.objects.get(f"{context.object.batoms.label}_cell")
            if obj is not None:
                return obj.data.vertices[i1 + 1].co[i2]
        return 0

    return getter


def set_cell(i1, i2):
    """Helper function to easily set cell-property."""

    def setter(self, value):
        context = bpy.context
        if context.object and context.object.batoms.type != "OTHER":
            obj = bpy.data.objects.get(f"{context.object.batoms.label}_cell")
            if obj is not None:
                obj.data.vertices[i1 + 1].co[i2] = value

    return setter


class CellProperties(bpy.types.PropertyGroup):
    def selected_batoms(self, context):
        if context.object.batoms.batom.flag:
            return context.object.batoms.batom.label
        return None

    def Callback_modify_pbc(self, context):
        cell = bpy.context.scene.batoms.cell
        pbc = cell.pbc
        modify_batoms_attr(self.selected_batoms(context), "pbc", pbc)

    width: FloatProperty(
        name="width",
        default=0.05,
        min=0.0,
        soft_max=0.1,
        description="width",
        get=get_attr("width", get_active_bpy_data_batoms("cell")),
        set=set_attr("width", set_module_attr("cell")),
    )

    color: FloatVectorProperty(
        name="color",
        size=4,
        subtype="COLOR",
        min=0,
        max=1,
        default=[0.2, 0.2, 0.2, 1],
        get=get_attr("color", get_active_bpy_data_batoms("cell")),
        set=set_attr("color", set_module_attr("cell")),
    )

    pbc: BoolProperty(
        name="pbc", default=True, description="pbc", update=Callback_modify_pbc
    )
    # getter/setter only used for float not float vector
    cell_a0: FloatProperty(
        name="a0",
        default=0,
        description="Cell",
        get=get_cell(0, 0),
        set=set_cell(0, 0),
    )
    cell_a1: FloatProperty(
        name="a1",
        default=0,
        description="Cell",
        get=get_cell(0, 1),
        set=set_cell(0, 1),
    )
    cell_a2: FloatProperty(
        name="a2",
        default=0,
        description="Cell",
        get=get_cell(0, 2),
        set=set_cell(0, 2),
    )
    cell_b0: FloatProperty(
        name="b0",
        default=0,
        description="Cell",
        get=get_cell(1, 0),
        set=set_cell(1, 0),
    )
    cell_b1: FloatProperty(
        name="b1",
        default=0,
        description="Cell",
        get=get_cell(1, 1),
        set=set_cell(1, 1),
    )
    cell_b2: FloatProperty(
        name="b2",
        default=0,
        description="Cell",
        get=get_cell(1, 2),
        set=set_cell(1, 2),
    )
    cell_c0: FloatProperty(
        name="c0",
        default=0,
        description="Cell",
        get=get_cell(2, 0),
        set=set_cell(2, 0),
    )
    cell_c1: FloatProperty(
        name="c1",
        default=0,
        description="Cell",
        get=get_cell(2, 1),
        set=set_cell(2, 1),
    )
    cell_c2: FloatProperty(
        name="c2",
        default=0,
        description="Cell",
        get=get_cell(2, 2),
        set=set_cell(2, 2),
    )

    transform_a: IntVectorProperty(
        name="a",
        default=(1, 0, 0, 0),
        size=4,
        min=-20,
        max=20,
        # subtype = "XYZ",
        description="Transform matrix in a axis",
    )
    transform_b: IntVectorProperty(
        name="b",
        default=(0, 1, 0, 0),
        size=4,
        min=-20,
        max=20,
        # subtype = "XYZ",
        description="Transform matrix in b axis",
    )
    transform_c: IntVectorProperty(
        name="c",
        default=(0, 0, 1, 0),
        size=4,
        min=-20,
        max=20,
        # subtype = "XYZ",
        description="Transform matrix in c axis",
    )

    boundary_a: FloatVectorProperty(
        name="a", default=(0, 1), size=2, description="boundary  in a axis"
    )
    boundary_b: FloatVectorProperty(
        name="b", default=(0, 1), size=2, description="boundary  in b axis"
    )
    boundary_c: FloatVectorProperty(
        name="c", default=(0, 1), size=2, description="boundary  in c axis"
    )


def modify_batoms_attr(name, key, value):
    if name is None:
        return
    batoms = Batoms(label=name)
    setattr(batoms, key, value)
    # batoms.select = True


def modify_transform(name, transform):
    if name is None:
        return
    batoms = Batoms(label=name)
    batoms.transform(transform)
    batoms.hide = True



import bpy
import pytest
from batoms import Batoms
from batoms.database.mp import mp_search


def test_database_materials_project():
    bpy.ops.batoms.delete()
    key = None
    if key:
        mp_search(key, "mp-2815")


def test_database_pubchem():
    from batoms.database.pubchem import pubchem_search

    bpy.ops.batoms.delete()
    batoms = pubchem_search("31423")
    assert batoms.label == "cid_31423"


def test_database_pymatgen():
    pytest.importorskip("pymatgen")
    from pymatgen.core.structure import Molecule
    from pymatgen.core import Lattice, Structure

    bpy.ops.batoms.delete()
    # molecule
    co = Molecule(["C", "O"], [[0.0, 0.0, 0.0], [0.0, 0.0, 1.2]])
    co = Batoms(label="co", from_pymatgen=co)
    # lattice
    fe = Structure(Lattice.cubic(2.8), ["Fe", "Fe"], [[0, 0, 0], [0.5, 0.5, 0.5]])
    fe = Batoms(label="fe", from_pymatgen=fe)
    assert fe.pbc == [True, True, True]



"""Definition of the polyhedras class.

This module defines the polyhedras object in the Batoms package.

"""

import bpy
from time import time
from batoms.utils.butils import object_mode, compareNodeType, get_node_by_name
from batoms.utils import string2Number
import numpy as np
from batoms.base.object import ObjectGN
from .setting import PolyhedraSettings
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


default_attributes = [
    {"name": "atoms_index1", "data_type": "INT"},
    {"name": "atoms_index2", "data_type": "INT"},
    {"name": "species_index", "data_type": "INT"},
    {"name": "face_species_index", "data_type": "INT", "domain": "FACE"},
    {"name": "show", "data_type": "INT"},
    {"name": "style", "data_type": "INT"},
]


default_GroupInput = [
    ["atoms_index1", "NodeSocketInt", "POINT"],
    ["atoms_index2", "NodeSocketInt", "POINT"],
    ["species_index", "NodeSocketInt", "POINT"],
    ["face_species_index", "NodeSocketInt", "FACE"],
    ["show", "NodeSocketBool", "POINT"],
    ["style", "NodeSocketInt", "POINT"],
]

default_polyhedra_datas = {
    "atoms_index1": np.ones(0, dtype=int),
    "atoms_index2": np.ones(0, dtype=int),
    "species_index": np.ones(0, dtype=int),
    "face_species_index": np.ones(0, dtype=int),
    "vertices": np.zeros((0, 3)),
    "offsets": np.zeros((0, 3)),
    "edges": [],
    "faces": [],
    "widths": np.ones(0, dtype=float),
    "shows": np.zeros(0, dtype=int),
}


class Polyhedra(ObjectGN):
    """Polyhedras Class

    A Polyhedras object is linked to this main collection in Blender.

    Parameters:

    label: str
        Name of the Bpolyhedras.
    species: str
        species of the atoms.
    positions: array
        positions
    locations: array
        The object’s origin location in global coordinates.
    element: str or list
        element of the atoms, list for fractional Occupancy
    segments: list of 2 Int
        Number of segments used to draw the UV_Sphere
        Default: [32, 16]
    subdivisions: Int
        Number of subdivision used to draw the ICO_Sphere
        Default: 2
    color_style: str
        "JMOL", "ASE", "VESTA"
    radii_style: str
        "covelent", "vdw", "ionic"
    shape: Int
        0, 1, or 2. ["UV_SPHERE", "ICO_SPHERE", "CUBE"]

    Examples:

    >>> from batoms.bond import Bbond
    >>> c = Bbond('C', [[0, 0, 0], [1.2, 0, 0]])

    """

    def __init__(
        self,
        label=None,
        polyhedra_datas=None,
        location=np.array([0, 0, 0]),
        batoms=None,
    ):
        #
        self.batoms = batoms
        self.label = label
        name = "polyhedra"
        ObjectGN.__init__(self, label, name)
        self.settings = PolyhedraSettings(self.label, batoms=batoms, polyhedras=self)
        flag = self.load()
        if not flag and polyhedra_datas is not None:
            self.build_object(polyhedra_datas)

    def build_object(self, datas, attributes={}):
        object_mode()
        """
        build child object and add it to main objects.
        """
        tstart = time()
        if len(datas["vertices"].shape) == 2:
            self._trajectory = {
                "vertices": np.array([datas["vertices"]]),
                "offsets": np.array([datas["offsets"]]),
            }
            vertices = datas["vertices"]
            offsets = datas["offsets"]
        elif len(datas["vertices"].shape) == 3:
            self._trajectory = {
                "vertices": datas["vertices"],
                "offsets": datas["offsets"],
            }
            vertices = datas["vertices"][0]
            offsets = datas["offsets"][0]
        else:
            raise Exception("Shape of vertices is wrong!")

        attributes.update(
            {
                "atoms_index1": datas["atoms_index1"],
                "atoms_index2": datas["atoms_index2"],
                "species_index": datas["species_index"],
                "face_species_index": datas["face_species_index"],
                "show": datas["shows"],
                # 'model_style': datas['model_styles'],
            }
        )
        name = self.obj_name
        self.delete_obj(name)
        mesh = bpy.data.meshes.new(name)
        mesh.from_pydata(vertices, datas["edges"], datas["faces"])
        mesh.update()
        obj = bpy.data.objects.new(name, mesh)
        # Add attributes
        for att in default_attributes:
            self.add_attribute(**att)
        self.settings.coll.objects.link(obj)
        obj.batoms.type = "POLYHEDRA"
        obj.batoms.label = self.label
        obj.parent = self.batoms.obj
        #
        name = "%s_polyhedra_offset" % self.label
        self.delete_obj(name)
        mesh = bpy.data.meshes.new(name)
        mesh.from_pydata(offsets, [], [])
        mesh.update()
        obj = bpy.data.objects.new(name, mesh)
        self.settings.coll.objects.link(obj)
        obj.hide_set(True)
        bpy.context.view_layer.update()
        self.set_attributes(attributes)
        self.init_geometry_node_modifier(default_GroupInput)
        self.build_geometry_node()
        self.set_trajectory()
        # self.assign_materials()
        self.update_geometry_node_material()
        obj.parent = self.obj
        logger.debug("polyhedras: build_object: {0:10.2f} s".format(time() - tstart))

    def assign_materials(self):
        # sort element by occu
        me = self.obj.data
        me.materials.clear()
        for sp in self.settings:
            me.materials.append(self.settings.materials[sp.species][0])

    def build_geometry_node(self):
        """
        Geometry node for everything!
        """
        from batoms.utils.butils import get_node_by_name

        tstart = time()
        links = self.gn_node_group.links
        nodes = self.gn_node_group.nodes
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        # ------------------------------------------------------------------
        # ------------------------------------------------------------------
        # calculate bond vector, length, rotation based on the index
        # Get four positions from batoms, bond and the second bond
        #  for high order bond plane
        ObjectBatoms = get_node_by_name(
            nodes, "%s_ObjectBatoms" % self.label, "GeometryNodeObjectInfo"
        )
        ObjectBatoms.inputs["Object"].default_value = self.batoms.obj
        PositionBatoms = get_node_by_name(
            nodes, "%s_PositionBatoms" % (self.label), "GeometryNodeInputPosition"
        )
        TransferBatoms = get_node_by_name(
            nodes, "%s_TransferBatoms" % (self.label), "GeometryNodeSampleIndex"
        )
        TransferBatoms.data_type = "FLOAT_VECTOR"
        links.new(ObjectBatoms.outputs["Geometry"], TransferBatoms.inputs[0])
        links.new(PositionBatoms.outputs["Position"], TransferBatoms.inputs[3])
        links.new(GroupInput.outputs[2], TransferBatoms.inputs["Index"])
        # ------------------------------------------------------------------
        # add positions with offsets
        # transfer offsets from object self.obj_o
        ObjectOffsets = get_node_by_name(
            nodes, "%s_ObjectOffsets" % (self.label), "GeometryNodeObjectInfo"
        )
        ObjectOffsets.inputs["Object"].default_value = self.obj_o
        PositionOffsets = get_node_by_name(
            nodes, "%s_PositionOffsets" % (self.label), "GeometryNodeInputPosition"
        )
        TransferOffsets = get_node_by_name(
            nodes, "%s_TransferOffsets" % self.label, "GeometryNodeSampleIndex"
        )
        TransferOffsets.data_type = "FLOAT_VECTOR"
        InputIndex = get_node_by_name(
            nodes, "%s_InputIndex" % self.label, "GeometryNodeInputIndex"
        )
        links.new(ObjectOffsets.outputs["Geometry"], TransferOffsets.inputs[0])
        links.new(PositionOffsets.outputs["Position"], TransferOffsets.inputs[3])
        links.new(InputIndex.outputs[0], TransferOffsets.inputs["Index"])
        # we need one add operation to get the positions with offset
        VectorAdd = get_node_by_name(
            nodes, "%s_VectorAdd" % (self.label), "ShaderNodeVectorMath"
        )
        VectorAdd.operation = "ADD"
        links.new(TransferBatoms.outputs[2], VectorAdd.inputs[0])
        links.new(TransferOffsets.outputs[2], VectorAdd.inputs[1])
        # set positions
        SetPosition = get_node_by_name(
            nodes, "%s_SetPosition" % self.label, "GeometryNodeSetPosition"
        )
        links.new(GroupInput.outputs["Geometry"], SetPosition.inputs["Geometry"])
        links.new(VectorAdd.outputs[0], SetPosition.inputs["Position"])
        links.new(SetPosition.outputs["Geometry"], GroupOutput.inputs["Geometry"])

        # find kinds by the names of species
        for sp in self.settings:
            self.add_geometry_node(sp.as_dict())
        logger.debug("Build geometry nodes for polyhedras: %s" % (time() - tstart))

    def add_geometry_node(self, sp):
        from batoms.utils.butils import get_node_by_name

        links = self.gn_node_group.links
        nodes = self.gn_node_group.nodes
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        previousNode = GroupOutput.inputs["Geometry"].links[0].from_socket
        # print(previousNode)
        # we need two compares for one species,
        # vertices
        CompareSpecies = get_node_by_name(
            nodes,
            "%s_CompareSpecies_%s_vertex" % (self.label, sp["species"]),
            compareNodeType,
        )
        CompareSpecies.operation = "EQUAL"
        CompareSpecies.inputs[1].default_value = string2Number(sp["species"])
        links.new(GroupInput.outputs[3], CompareSpecies.inputs[0])
        # face
        CompareSpeciesFace = get_node_by_name(
            nodes,
            "%s_CompareSpecies_%s_face" % (self.label, sp["species"]),
            compareNodeType,
        )
        CompareSpeciesFace.operation = "EQUAL"
        CompareSpeciesFace.inputs[1].default_value = string2Number(sp["species"])
        links.new(GroupInput.outputs[4], CompareSpeciesFace.inputs[0])
        #
        setMaterialIndex = get_node_by_name(
            nodes,
            "%s_setMaterialIndex_%s" % (self.label, sp["species"]),
            "GeometryNodeSetMaterialIndex",
        )
        setMaterialIndex.inputs[2].default_value = self.settings.materials[
            sp["species"]
        ][1]
        links.new(previousNode, setMaterialIndex.inputs["Geometry"])
        links.new(CompareSpeciesFace.outputs[0], setMaterialIndex.inputs["Selection"])
        links.new(setMaterialIndex.outputs["Geometry"], GroupOutput.inputs["Geometry"])

    def set_realize_instances(self, realize_instances):
        """Make instancing object real
        # TODO: add make real to geometry node
        """
        #
        nodes = self.gn_node_group.nodes
        i = len(self.settings.bpy_setting) - 1
        sp = self.settings.bpy_setting[i]
        setMaterialIndex = get_node_by_name(
            nodes,
            "%s_setMaterialIndex_%s" % (self.label, sp["species"]),
            "GeometryNodeSetMaterialIndex",
        )

        RealizeInstances = get_node_by_name(
            nodes, "%s_RealizeInstances" % self.label, "GeometryNodeRealizeInstances"
        )
        if not realize_instances:
            # switch off
            if len(RealizeInstances.outputs[0].links) > 0:
                link = RealizeInstances.outputs[0].links[0]
                self.gn_node_group.links.remove(link)
            self.gn_node_group.links.new(
                setMaterialIndex.outputs[0], nodes[1].inputs[0]
            )
        else:
            self.gn_node_group.links.new(
                setMaterialIndex.outputs[0], RealizeInstances.inputs[0]
            )
            self.gn_node_group.links.new(
                RealizeInstances.outputs[0], nodes[1].inputs[0]
            )
        self.gn_node_group.update_tag()

    def update(
        self,
    ):
        """Draw polyhedras.
        calculate bond in all farmes, and save
        get the max number of polyhedras for each pair
        draw the polyhedras
        add shape key
        the extra polyhedras use the find bond data.
        """
        # if not self.bondlist:
        object_mode()
        # clean_coll_objects(self.coll, 'bond')
        positions = self.batoms.positions
        trajectory = self.batoms.get_trajectory()["positions"]
        arrays = self.batoms.arrays
        show = arrays["show"].astype(bool)
        species = arrays["species"]
        # trajectory_boundary = self.batoms.get_trajectory(self.batoms.batoms_boundary)
        # trajectory_search = self.batoms.get_trajectory(self.batoms.batoms_search)
        if self.batoms.nframe == 0:
            trajectory = np.array([positions])
        nframe = len(trajectory)
        polyhedra_datas = {}
        tstart = time()
        for f in range(nframe):
            logger.debug("update polyhedra: {}".format(f))
            positions = trajectory[f, show, :]
            bondlists = self.bondlists
            if len(bondlists) == 0:
                self.batoms.bond.update()
                bondlists = self.batoms.bond.bondlists
            polyhedra_kinds = self.calc_polyhedra_data(
                bondlists, species, positions, arrays["model_style"][show]
            )
            if f == 0:
                polyhedra_datas = polyhedra_kinds

        if len(polyhedra_datas) == 0:
            return
        self.set_arrays(polyhedra_datas)
        # self.coll.objects.link(bb.obj)
        # bpy.data.collections['Collection'].objects.unlink(bb.obj)
        # bb.set_trajectory()
        # bpy.context.scene.frame_set(self.batoms.nframe)
        logger.debug("draw polyhedra: {0:10.2f} s".format(time() - tstart))

    def update_geometry_node_material(self):
        """
        Make sure all species has a geometry node flow
        and the material are updated.
        """
        tstart = time()
        for sp in self.settings:
            sp = sp.as_dict()
            # update material in geometry node
            self.settings.build_materials(sp)
        self.assign_materials()
        for i in range(len(self.settings.bpy_setting)):
            sp = self.settings.bpy_setting[i]
            setMaterialIndex = get_node_by_name(
                self.gn_node_group.nodes,
                "%s_setMaterialIndex_%s" % (self.label, sp["species"]),
                "GeometryNodeSetMaterialIndex",
            )
            setMaterialIndex.inputs[2].default_value = i
        logger.debug("update polyhedra materials: %s" % (time() - tstart))

    @property
    def obj_o(self):
        return self.get_obj_o()

    def get_obj_o(self):
        name = "%s_polyhedra_offset" % self.label
        obj_o = bpy.data.objects.get(name)
        if obj_o is None:
            raise KeyError("%s object is not exist." % name)
        return obj_o

    def get_arrays(self):
        """ """
        object_mode()
        # tstart = time()
        arrays = self.attributes
        arrays.update(
            {
                "positions": self.positions[0],
                "offsets": self.positions[1],
            }
        )
        # print('get_arrays: %s'%(time() - tstart))
        return arrays

    @property
    def bondlists(self):
        return self.get_bondlists()

    def get_bondlists(self):
        bondlists = self.batoms.bond.bondlists
        return bondlists

    def set_arrays(self, arrays):
        """ """
        attributes = self.attributes
        # same length
        if len(arrays["vertices"]) == len(attributes["show"]):
            self.positions = arrays["vertices"]
            self.offsets = arrays["offsets"]
            self.set_attributes({"atoms_index1": arrays["atoms_index1"]})
            self.set_attributes({"atoms_index2": arrays["atoms_index2"]})
            self.set_attributes({"species_index": arrays["species_index"]})
            self.set_attributes({"face_species_index": arrays["face_species_index"]})
            self.update_geometry_node_material()
        else:
            # add or remove vertices, rebuild the object
            self.build_object(arrays)

    @property
    def obj(self):
        return self.get_obj()

    def get_obj(self):
        obj = bpy.data.objects.get(self.obj_name)
        if obj is None:
            self.build_object(default_polyhedra_datas)
        return obj

    @property
    def offsets(self):
        return self.get_offsets()

    def get_offsets(self):
        """
        using foreach_get and foreach_set to improve performance.
        """
        n = len(self)
        offsets = np.empty(n * 3, dtype=np.float64)
        self.obj_o.data.vertices.foreach_get("co", offsets)
        return offsets.reshape((n, 3))

    @offsets.setter
    def offsets(self, offsets):
        self.set_offsets(offsets)

    def set_offsets(self, offsets):
        """
        Set global offsets to local vertices
        """
        object_mode()
        n = len(self.obj_o.data.vertices)
        if len(offsets) != n:
            raise ValueError("offsets has wrong shape %s != %s." % (len(offsets), n))
        if n == 0:
            return
        offsets = offsets.reshape((n * 3, 1))
        self.obj_o.data.shape_keys.key_blocks[0].data.foreach_set("co", offsets)
        self.obj_o.data.update()

    def get_trajectory(self):
        """ """
        trajectory = {}
        trajectory["vertices"] = self.get_obj_trajectory(self.obj)
        trajectory["offsets"] = self.get_obj_trajectory(self.obj_o)
        return trajectory

    def set_trajectory(self, trajectory=None, frame_start=0):
        if trajectory is None:
            trajectory = self._trajectory
        nframe = len(trajectory)
        if nframe == 0:
            return
        name = "%s_polyhedra" % (self.label)
        obj = self.obj
        self.set_shape_key(name, obj, trajectory["vertices"], frame_start=frame_start)
        #
        name = "%s_polyhedra_offset" % (self.label)
        obj = self.obj_o
        self.set_shape_key(name, obj, trajectory["offsets"], frame_start=frame_start)

    def __getitem__(self, index):
        """Return a subset of the polyhedras.
        # TODO
        """
        pass

    def __setitem__(self, index, value):
        """set a subset of the polyhedras.
        # TODO
        """
        pass

    def calc_polyhedra_data(self, bondlists, species, positions, model_styles):
        """ """
        from ase.data import chemical_symbols
        from scipy.spatial import ConvexHull

        tstart = time()
        chemical_symbols = np.array(chemical_symbols)
        # find bonds contribute to polyhedra
        indices = bondlists[:, 8].astype(bool)
        bondlists = bondlists[indices]
        # maxinum number of poly
        npoly = len(bondlists)
        if npoly == 0:
            return default_polyhedra_datas
        # properties
        npoly = npoly * 12
        atoms_index1 = np.zeros(npoly, dtype=int)  # 0, 1, 2
        atoms_index2 = np.zeros(npoly, dtype=int)  # 0, 1, 2
        face_species_index = np.zeros(npoly, dtype=int)  # 0, 1, 2
        widths = np.ones(npoly, dtype=float)
        species_index = np.zeros(npoly, dtype=int)
        # ------------------------------------
        # offsets10 = bondlists[:, 2:5]
        offsets20 = bondlists[:, 5:8]
        indices10 = bondlists[:, 0].astype(int)
        indices20 = bondlists[:, 1].astype(int)
        positions20 = positions[indices20] + offsets20
        # ---------------------------------------------
        vertices = []
        offsets = []
        edges = []
        faces = []
        nv = 0
        nf = 0
        speciesarray = species[indices10]
        for poly in self.settings:
            # find center atoms == species
            spis = np.where(speciesarray == poly.species)[0]
            indices1 = indices10[spis]
            bondlist1 = bondlists[spis]
            # center atoms is define by i and the offset
            atom_centers = bondlist1[:, [0, 2, 3, 4]]
            positions21 = positions20[spis]
            # find indices of center atoms
            u, indices = np.unique(atom_centers, axis=0, return_inverse=True)
            indices = np.append(indices, len(indices1))
            n = len(u)
            # print(n)
            for i in range(n):
                if model_styles[int(u[i, 0])] != 2:
                    continue
                mask = np.where(indices == i)[0]
                vertices1 = positions21[mask]
                dnv = len(vertices1)
                if dnv >= 4:
                    # search face for polyhedra
                    hull = ConvexHull(vertices1)
                    face = hull.simplices
                    # tri = Delaunay(vertices1)
                    # face = tri.simplices
                    dnf = len(face)
                    face = face + nv
                    edge = []
                    for f in face:
                        edge.append([f[0], f[1]])
                        edge.append([f[0], f[2]])
                        edge.append([f[1], f[2]])
                    edges = edges + list(edge)
                    faces = faces + list(face)
                    vertices.extend(vertices1)
                    offsets.extend(bondlist1[mask, 5:8])
                    nv1 = nv + dnv
                    nf1 = nf + dnf
                    atoms_index1[nv:nv1] = bondlist1[mask, 0]
                    atoms_index2[nv:nv1] = bondlist1[mask, 1]
                    widths[nv:nv1] = poly.width
                    species_index[nv:nv1] = string2Number(poly.species)
                    face_species_index[nf:nf1] = string2Number(poly.species)
                    nv = nv1
                    nf = nf1
        n = len(vertices)
        if n == 0:
            datas = default_polyhedra_datas
        else:
            nf = len(faces)
            shows = np.ones(n, dtype=int)
            datas = {
                "atoms_index1": atoms_index1[0:n],
                "atoms_index2": atoms_index2[0:n],
                "species_index": species_index[0:n],
                "face_species_index": face_species_index[0:nf],
                "vertices": np.array(vertices),
                "offsets": np.array(offsets),
                "widths": widths[0:n],
                "edges": edges,
                "faces": faces,
                "shows": shows,
                "model_styles": model_styles,
            }
        # print('datas: ', datas)
        logger.debug("calc_polyhedra_data: {0:10.2f} s".format(time() - tstart))
        return datas

    @property
    def setting(self):
        from batoms.utils import deprecated

        """setting object."""
        deprecated(
            '"setting" will be deprecated in the furture, please use "settings".'
        )
        return self.settings

    def as_dict(self):
        """ """
        data = {"array": None}
        if len(self) > 1:
            data["array"] = dict(self.arrays)
        data["settings"] = self.settings.as_dict()
        return data



import bmesh
from bpy.props import BoolProperty, FloatProperty, IntVectorProperty, StringProperty
from batoms import Batoms
from batoms.ops.base import OperatorBatoms


class LatticePlaneAdd(OperatorBatoms):
    bl_idname = "plane.lattice_plane_add"
    bl_label = "Add Lattice Plane"
    bl_description = "Add Lattice Plane to a Batoms"

    indices: IntVectorProperty(name="Miller indices", size=3, default=[0, 0, 1])

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=context.object.batoms.label)
        batoms.lattice_plane.settings.add(self.indices)
        context.view_layer.objects.active = obj
        self.report(
            {"INFO"},
            "Add lattice plane ({}{}{}).".format(
                self.indices[0], self.indices[1], self.indices[2]
            ),
        )
        return {"FINISHED"}


class LatticePlaneRemove(OperatorBatoms):
    bl_idname = "plane.lattice_plane_remove"
    bl_label = "Remove Lattice Plane"
    bl_description = "Remove Lattice Plane to a Batoms"

    name: StringProperty(
        name="name", default="1-1-1", description="Name of Lattice Plane to be removed"
    )

    all: BoolProperty(
        name="all", default=False, description="Remove all Lattice Planes"
    )

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.lattice_plane.settings.remove((self.name))
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class LatticePlaneDraw(OperatorBatoms):
    bl_idname = "plane.lattice_plane_draw"
    bl_label = "Draw Lattice Plane"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Draw Lattice Plane"

    name: StringProperty(
        name="name", default="ALL", description="Name of Lattice Plane to be drawed"
    )

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.lattice_plane.draw(self.name)
        context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class LatticePlaneModify(OperatorBatoms):
    bl_idname = "plane.lattice_plane_modify"
    bl_label = "Modify lattice_plane"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Modify lattice_plane"

    key: StringProperty(
        name="key", default="style", description="Replaced by this species"
    )

    slice: BoolProperty(
        name="slice",
        default=False,
    )
    boundary: BoolProperty(
        name="boundary",
        default=False,
    )
    distance: FloatProperty(
        name="distance", description="Distance from origin", default=1
    )

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type == "BOND" and obj.mode == "EDIT"
        else:
            return False

    def execute(self, context):
        obj = context.object
        data = obj.data
        bm = bmesh.from_edit_mesh(data)
        v = [s.index for s in bm.select_history if isinstance(s, bmesh.types.BMVert)]
        batoms = Batoms(label=obj.batoms.label)
        for i in v:
            setattr(batoms.bonds[i], self.key, getattr(self, self.key))
        # batoms.draw()
        context.view_layer.objects.active = obj
        return {"FINISHED"}



import bpy
from batoms.batoms import Batoms
from ase.build import molecule, bulk
from batoms.bio.bio import read
from time import time
import numpy as np

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_bond(c2h6so):
    c2h6so.model_style = 1
    c2h6so.show = [0, 0, 1, 1, 1, 1, 1, 1, 1, 1]
    c2h6so.model_style = 1
    c2h6so.bond[0].order = 2
    c2h6so.bond.settings["C-H"].order = 2
    c2h6so.model_style = 1
    c2h6so.bond[0].order = 2


def test_new_species(ch4):
    """Add a new species.
    Should create a new bond pair correctly."""
    ch4.model_style = 1
    ch4.replace([1], "O")
    ch4.model_style = 1
    depsgraph = bpy.context.evaluated_depsgraph_get()
    # Get Geometry Node Instances
    eval_obj = ch4.obj.evaluated_get(depsgraph)
    insts = [
        inst
        for inst in depsgraph.object_instances
        if inst.is_instance and inst.parent == eval_obj
    ]
    # there are 5 atoms and 4 bonds
    assert len(insts) == 9


def test_settings(c2h6so):
    """key search"""
    # species tuple
    assert c2h6so.bond.settings.find(("C", "H")) is not None
    # species list
    assert c2h6so.bond.settings.find(["C", "H"]) is not None
    # name str
    assert c2h6so.bond.settings.find("C-H") is not None
    # remove by tuple
    c2h6so.bond.settings.remove(("C", "H"))
    assert c2h6so.bond.settings.find(("C", "H")) is None
    # add by tuple
    c2h6so.bond.settings.add(("C", "H"))
    assert c2h6so.bond.settings.find(("C", "H")) is not None
    # remove by str
    c2h6so.bond.settings.remove("C-H")
    assert c2h6so.bond.settings.find(("C", "H")) is None


def test_color(c2h6so):
    c2h6so.model_style = 1
    assert np.isclose(
        c2h6so.bond.settings.instancers["C-H"]["1_1"]
        .data.materials[1]
        .diffuse_color[:],
        np.array([1, 1, 1, 1]),
    ).all()
    c2h6so.bond.settings["C-H"].color1 = [0, 1, 0, 1]
    c2h6so.bond.settings["C-H"].color2 = [0, 1, 1, 1]
    assert np.isclose(
        c2h6so.bond.settings.instancers["C-H"]["1_1"]
        .data.materials[1]
        .diffuse_color[:],
        np.array([0, 1, 1, 1]),
    ).all()


def test_bond_high_order():
    from ase.build import molecule

    bpy.ops.batoms.delete()
    c6h6 = Batoms("c6h6", from_ase=molecule("C6H6"))
    c6h6.model_style = 1
    c6h6.bond[0].order = 2
    c6h6.bond[2].order = 2
    c6h6.bond[5].order = 2
    c6h6.bond.settings["C-C"].order = 2


def test_bond_performance(h2o):
    h2o.cell = [3, 3, 3]
    h2o.pbc = True
    h2o = h2o * [10, 10, 10]
    tstart = time()
    h2o.model_style = 1
    t = time() - tstart
    assert t < 5


def test_bond_add():
    bpy.ops.batoms.delete()
    au = bulk("Au")
    au = Batoms("au", from_ase=au)
    au = au * [2, 2, 2]
    assert len(au.bond.settings) == 0
    au.bond.settings.add(("Au", "Au"))
    assert len(au.bond.settings) == 1


def test_bond_search_bond_0(tio2):
    tio2.boundary = 0.01
    tio2.model_style = 1
    depsgraph = bpy.context.evaluated_depsgraph_get()
    # Get Geometry Node Instances
    eval_obj = tio2.obj.evaluated_get(depsgraph)
    insts = [
        inst
        for inst in depsgraph.object_instances
        if inst.is_instance and inst.parent == eval_obj
    ]
    # there are 6 atoms and 54 bonds
    assert len(insts) == 60
    # change search bond settings
    tio2.bond.settings[("Ti", "O")].search = 0
    tio2.model_style = 1
    tio2.boundary = 0.01
    tio2.model_style = 1
    assert len(tio2.bond.bondlists) == 14
    eval_obj = tio2.obj.evaluated_get(depsgraph)
    insts = [
        inst
        for inst in depsgraph.object_instances
        if inst.is_instance and inst.parent == eval_obj
    ]
    assert len(insts) == 20
    tio2.bond.show_search = True
    if use_cycles:
        set_cycles_res(tio2)
    tio2.get_image([1, -0.3, 0.1], output="bond_search_0.png", **extras)


def test_bond_search_bond_1():
    bpy.ops.batoms.delete()
    mol = read("../tests/datas/anthraquinone.cif")
    mol.boundary = 0.01
    mol.model_style = 1
    mol.bond.show_search = True
    if use_cycles:
        set_cycles_res(mol)
    mol.get_image([1, -0.3, 0.1], output="anthraquinone.png", **extras)


def test_bond_search_bond_2():
    bpy.ops.batoms.delete()
    mof = read("../tests/datas/mof-5.cif")
    mof.boundary = 0.01
    mof.bond.settings[("Zn", "O")].polyhedra = True
    mof.model_style = 1
    if use_cycles:
        set_cycles_res(mof)
    mof.get_image([0, 1, 0], output="mof-5.png", **extras)


def test_bond_search_bond_3(tio2):
    tio2.boundary = 0.01
    tio2.model_style = 1
    # assert len(tio2.bond.search_bond) < 20


def test_hydrogen_bond():
    bpy.ops.batoms.delete()
    ch3oh = Batoms(label="ch3oh", from_ase=molecule("CH3OH"))
    ch3oh.bond.settings[("H", "O")].min = 2.0
    ch3oh.bond.settings[("H", "O")].max = 3.0
    ch3oh.bond.settings[("H", "O")].style = "2"
    ch3oh.model_style = 1
    if use_cycles:
        set_cycles_res(ch3oh)
    ch3oh.get_image([1, 0, 0], output="bond-hb.png", **extras)


def test_bond_reload(tio2):
    """save to blend file and reload"""
    import os

    tio2.boundary = 0.01
    tio2.model_style = 1
    tio2.bond.show_search = True
    cwd = os.getcwd()
    filepath = os.path.join(cwd, "test.blend")
    bpy.ops.wm.save_as_mainfile(filepath=filepath)
    bpy.ops.batoms.delete()
    bpy.ops.wm.open_mainfile(filepath=filepath)
    tio2 = Batoms("tio2")
    assert len(tio2.bond.search_bond) == 43



"""
"""
import bpy
from bpy_extras import view3d_utils
import bmesh
from batoms import Batoms
import blf
import gpu
from gpu_extras.batch import batch_for_shader
import numpy as np


def interpolate_2points(points, dash):
    v = points[1] - points[0]
    d = np.linalg.norm(v)
    n = max(1, int(d / dash))
    v = v / n
    d = np.linspace(0, n, n + 1)
    vertices = points[0] + d[:, None] * v
    return vertices.tolist()


def draw_callback_text(self, context):
    # get the context arguments
    if len(self._positions) == 0:
        return
    region = context.region
    rv3d = context.region_data
    # coord = event.mouse_region_x, event.mouse_region_y
    font_id = 0  # XXX, need to find out how best to get this.
    coord = np.mean(self._positions, axis=0)
    coord_2d = view3d_utils.location_3d_to_region_2d(region, rv3d, coord, default=None)
    # draw some text
    blf.position(font_id, coord_2d[0], coord_2d[1] + 10, 0)
    blf.size(font_id, 20, 72)
    blf.draw(font_id, self.results)


def draw_callback_line(self, context, dash=0.1):
    # 50% alpha, 2 pixel width line
    # get the context arguments
    if len(self._positions) == 0:
        return
    n = int(len(self._positions) / 2)
    vertices = []
    for i in range(n):
        vertices.extend(
            interpolate_2points(
                [self._positions[2 * i], self._positions[2 * i + 1]], dash
            )
        )
    shader = gpu.shader.from_builtin("3D_UNIFORM_COLOR")
    # gpu.state.blend_set('ALPHA')
    gpu.state.line_width_set(2.0)
    batch = batch_for_shader(shader, "LINES", {"pos": vertices})
    shader.bind()
    shader.uniform_float("color", (0.0, 0.0, 0.0, 1))
    batch.draw(shader)

    # restore opengl defaults
    gpu.state.line_width_set(1.0)
    gpu.state.blend_set("NONE")


class MeasureButton(bpy.types.Operator):
    """Records the selection order while running and
    when finished with ESC"""

    bl_idname = "batoms.measure"
    bl_label = "Measurement"
    # bl_options = {'REGISTER', 'UNDO'} # this is needed so we actally get a UI

    results = ""

    dash = 0.1

    @classmethod
    def poll(cls, context):
        return context.mode in {"EDIT_MESH"}

    def modal(self, context, event):
        context.area.tag_redraw()
        if event.type in {"RIGHTMOUSE"}:
            obj = context.object
            data = obj.data
            bm = bmesh.from_edit_mesh(data)
            v = [
                s.index for s in bm.select_history if isinstance(s, bmesh.types.BMVert)
            ]
            batoms = Batoms(label=obj.batoms.label)
            results = []
            nv = len(v)
            # self.positions = [tuple(batoms[0].position)]
            for i in range(1, nv):
                self.positions.append(batoms[v[i - 1]].position)
                self.positions.append(batoms[v[i]].position)
            if len(v) == 1:
                measurement_type = "%s: " % batoms[v[0]].species
                results = batoms[v[0]].position
            elif len(v) == 2:
                results = batoms.get_distances(v[0], [v[1]])
                measurement_type = ""
            elif len(v) == 3:
                results = batoms.get_angle(v[0], v[1], v[2])
                measurement_type = ""
            elif len(v) == 4:
                results = batoms.get_dihedral(v[0], v[1], v[2], v[3])
                measurement_type = ""
            else:
                measurement_type = "More than four atoms are selected. Unsupported."
            # update measurement value
            results.shape = (-1,)
            results = [str(round(float(i), 2)) for i in results]
            results = measurement_type + " ".join(results)
            # self.report({'INFO'}, results)
            bpy.ops.object.mode_set(mode="EDIT")
            self.results = results
            self._positions = self.positions
            self.positions = []
            return {"RUNNING_MODAL"}
        elif event.type in {"ESC"}:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_line, "WINDOW")
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_text, "WINDOW")
            return {"CANCELLED"}
        return {"PASS_THROUGH"}

    def invoke(self, context, event):
        if context.area.type == "VIEW_3D":
            # the arguments we pass the the callback
            args = (self, context, self.dash)
            # Add the region OpenGL drawing callback
            # draw in view space with 'POST_VIEW' and 'PRE_VIEW'
            self._handle_line = bpy.types.SpaceView3D.draw_handler_add(
                draw_callback_line, args, "WINDOW", "POST_VIEW"
            )
            args = (self, context)
            self._handle_text = bpy.types.SpaceView3D.draw_handler_add(
                draw_callback_text, args, "WINDOW", "POST_PIXEL"
            )
            self.results = ""
            self.positions = []
            self._positions = []
            context.window_manager.modal_handler_add(self)
            return {"RUNNING_MODAL"}
        else:
            self.report({"WARNING"}, "View3D not found, cannot run operator")
            return {"CANCELLED"}



"""
"""
import numpy as np
from time import time
from ase.geometry import wrap_positions
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


def RemovePbc(species0, positions0, cell, pbc, cutoffs):
    """
    convert atoms with pbc to non-pbc by adding boundary
    atoms within max cutoff
    """
    maxCutoff = np.array([x for x in cutoffs.values()]).max()
    boundary = np.ones((3, 2)) * maxCutoff
    array = build_boundary(species0, positions0, cell, pbc, boundary, include_self=True)
    return array


def bondlist_kdtree(
    quantities, species0, positions0, cell, pbc, setting, self_interaction=False
):
    """
    return

    i: index1
    j: index2
    k: search bond style
    """
    cutoffs = {}
    for pair, b in setting.items():
        cutoffs[pair] = [b["min"], b["max"]]
    natom = len(positions0)
    # orignal atoms
    array1 = {
        "positions": positions0,
        "species": species0,
        "indices": np.arange(natom),
        "offsets": np.zeros((natom, 3)),
    }
    # atoms added with boundary
    array2 = RemovePbc(species0, positions0, cell, pbc, cutoffs)
    bonddatas = primitive_neighbor_kdtree(array1, array2, cutoffs)
    # build bondlist
    i = []
    j = []
    k = []
    p = []
    t = []
    for pair, data in bonddatas.items():
        if len(data) == 0:
            continue
        i1 = []
        j1 = []
        for i2, j2 in data.items():
            n = len(j2)
            i1.extend([i2] * n)
            j1.extend(j2)
        # remove bothways for same species, e.g. ('C', 'C')
        if pair[0] == pair[1]:
            i1 = np.array(i1)
            j1 = np.array(j1)
            # wrong, offset1 could be non zero
            mask = np.where(
                (array1["indices"][i1] > array2["indices"][j1]), False, True
            )
            i1 = i1[mask]
            j1 = j1[mask]
            i1 = list(i1)
            j1 = list(j1)
        n = len(i1)
        k1 = [setting[pair]["search"]] * n
        p1 = [setting[pair]["polyhedra"]] * n
        t1 = [setting[pair]["type"]] * n
        i.extend(i1)
        j.extend(j1)
        k.extend(k1)
        p.extend(p1)
        t.extend(t1)
    # offsets
    offsets_i = array1["offsets"][i]
    offsets_j = array2["offsets"][j]
    distance_vector = array1["positions"][i] - array2["positions"][j]
    distances = np.sqrt(np.sum(distance_vector * distance_vector, axis=1))
    #
    i = array1["indices"][i]
    j = array2["indices"][j]
    k = np.array(k)
    p = np.array(p)
    t = np.array(t)
    # Remove all self-interaction.
    if not self_interaction:
        mask = np.where(
            (i == j) & ((array1["offsets"][i] == array2["offsets"][j]).all(axis=1)),
            False,
            True,
        )
        # print(mask)
        i = i[mask]
        j = j[mask]
        k = k[mask]
        p = p[mask]
        t = t[mask]
        distances = distances[mask]
        offsets_i = offsets_i[mask]
        offsets_j = offsets_j[mask]
    #
    retvals = []
    for q in quantities:
        if q == "i":
            retvals += [i]
        elif q == "j":
            retvals += [j]
        elif q == "k":
            retvals += [k]
        elif q == "p":
            retvals += [p]
        elif q == "t":
            retvals += [t]
        elif q == "d":
            retvals += [distances]
        elif q == "S":
            retvals += [offsets_j]
        else:
            raise ValueError("Unsupported quantity specified.")
    if len(retvals) == 1:
        return retvals[0]
    else:
        return tuple(retvals)


def neighbor_kdtree(species0, positions0, cell, pbc, cutoffs):
    """
    wrap to pbc structure

    """
    natom = len(positions0)
    # orignal atoms
    array1 = {
        "positions": positions0,
        "species": species0,
        "indices": np.arange(natom),
        "offsets": np.zeros((natom, 3)),
    }
    # atoms added with boundary
    array2 = RemovePbc(species0, positions0, cell, pbc, cutoffs)
    bonddatas = primitive_neighbor_kdtree(array1, array2, cutoffs)
    # bondlists = neighbor_kdtree(quantities,
    # array1, array2, cutoffs, self_interaction)
    return bonddatas


def primitive_neighbor_kdtree(array1, array2, cutoffs, parallel=1):
    """non pbc
    build bond lists between atoms1 and atoms2.
    In the non-pbc case: atoms1 and atoms2 could be the same.
    In pbc case, atoms2 is atoms1 + boundary atoms
    """
    from scipy.spatial import KDTree

    tstart = time()
    bonddatas = {}
    for pair, cutoff in cutoffs.items():
        bonddatas[pair] = {}
        indices_i = np.where(array1["species"] == pair[0])[0]
        indices_j = np.where(array2["species"] == pair[1])[0]
        if len(indices_i) == 0 or len(indices_j) == 0:
            continue
        indices_min = None
        # qurey the smaller array is faster, flip is needed
        p1 = array1["positions"][indices_i]
        p2 = array2["positions"][indices_j]
        # find max
        tree = KDTree(p2)
        indices_max = tree.query_ball_point(p1, r=cutoff[1], workers=parallel)
        # find min
        if cutoff[0] > 1e-6:
            indices_min = tree.query_ball_point(p1, r=cutoff[0], workers=parallel)
        #
        n = len(p1)
        if indices_min is not None:
            for k in range(n):
                indices_max[k] = list(set(indices_max[k]) - set(indices_min[k]))
                m = len(indices_max[k])
                if m == 0:
                    continue
                bonddatas[pair][indices_i[k]] = indices_j[indices_max[k]]
        else:
            for k in range(n):
                m = len(indices_max[k])
                if m == 0:
                    continue
                bonddatas[pair][indices_i[k]] = indices_j[indices_max[k]]
    logger.debug("Build bondlist: {:1.2f}".format(time() - tstart))
    # print(bonddatas)
    return bonddatas


def cellPlanes(cell, origin=np.array([0, 0, 0])):
    """
    build six planes for the six faces of a cell
    """
    c1, c2, c3 = cell
    normals = np.zeros((6, 3))
    points = np.zeros((6, 3))
    normals[0] = np.cross(c2, c3)
    normals[1] = -normals[0]
    normals[2] = np.cross(c3, c1)
    normals[3] = -normals[2]
    normals[4] = np.cross(c1, c2)
    normals[5] = -normals[4]
    norm = np.linalg.norm(normals, axis=1) + 1e-8
    normals = normals / norm[:, None]
    points[0] = origin
    points[1] = origin + c1
    points[2] = origin
    points[3] = origin + c2
    points[4] = origin
    points[5] = origin + c3
    planes = []
    for i in range(3):
        plane = [
            {"normals": normals[2 * i], "points": points[2 * i]},
            {"normals": normals[2 * i + 1], "points": points[2 * i + 1]},
        ]
        planes.append(plane)
    return planes


def pointCellDistance(points, cell, origin=np.zeros(3)):
    """
    get distance between point and the six faces of a cell.
    """
    planes = cellPlanes(cell, origin=origin)
    distances = []
    npoint = len(points)
    distances = np.zeros((3, 2, npoint))
    for i in range(3):
        for j in range(2):
            vec = points - planes[i][j]["points"]
            distances[i, j, :] = np.dot(vec, planes[i][j]["normals"])
    return distances


def build_boundary(species, positions, cell, pbc, boundary, include_self=False):
    """
    find region outside cell for periodic caculation, supercell

    todo: boundary > 1
    """
    from functools import reduce

    # tstart = time()
    wraped_positions = wrap_positions(positions, cell, pbc=pbc)
    distances = pointCellDistance(wraped_positions, cell)
    natom = len(positions)
    # find atoms close to cell face with distance of r
    indices = [[[], []], [[], []], [[], []]]
    nb = 0
    for i in range(3):
        if not pbc[i]:
            continue
        ind0 = np.where((distances[i, 0, :] < boundary[i, 1]))[0]
        ind1 = np.where((distances[i, 1, :] < boundary[i, 0]))[0]
        indices[i] = [ind0, ind1]
        nb += len(ind0) + len(ind1)
    # print(indices)
    # init
    # repeat the scaled_positions so that it completely covers the boundary
    positions_b = np.zeros((27 * nb, 3))
    offsets_b = np.zeros((27 * nb, 3))
    indices_b = np.zeros(27 * nb, dtype=int)
    species_b = np.zeros(27 * nb, dtype="U20")
    # build face
    offset = [1, -1]
    nb1 = 0
    nb2 = 0
    for c in range(3):
        if not pbc[i]:
            continue
        for i in range(2):
            nb2 = nb1 + len(indices[c][i])
            indices_b[nb1:nb2] = indices[c][i]
            positions_b[nb1:nb2] = positions[indices[c][i]]
            offsets_b[nb1:nb2][:, c] = offset[i]
            nb1 = nb2
    # build edge
    for c in [[0, 1], [0, 2], [1, 2]]:
        if not pbc[c[0]] or not pbc[c[1]]:
            continue
        for i in range(2):
            for j in range(2):
                indices_j = np.intersect1d(indices[c[0]][i], indices[c[1]][j])
                nb2 = nb1 + len(indices_j)
                indices_b[nb1:nb2] = indices_j
                positions_b[nb1:nb2] = positions[indices_j]
                offsets_b[nb1:nb2][:, c[0]] = offset[i]
                offsets_b[nb1:nb2][:, c[1]] = offset[j]
                nb1 = nb2
    # build corner
    if pbc[0] and pbc[1] and pbc[2]:
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    indices3 = reduce(
                        np.intersect1d, (indices[0][i], indices[1][j], indices[2][k])
                    )
                    nb2 = nb1 + len(indices3)
                    indices_b[nb1:nb2] = indices3
                    positions_b[nb1:nb2] = positions[indices3]
                    offsets_b[nb1:nb2][:, 0] = offset[i]
                    offsets_b[nb1:nb2][:, 1] = offset[j]
                    offsets_b[nb1:nb2][:, 2] = offset[k]
                    nb1 = nb2
    positions_b = positions_b[0:nb2]
    indices_b = indices_b[0:nb2]
    species_b = species[indices_b]
    offsets_b = offsets_b[0:nb2]
    positions_b = positions_b + np.dot(offsets_b, cell)
    if include_self:
        positions_b = np.append(positions, positions_b, axis=0)
        indices_b = np.append(np.arange(natom), indices_b)
        species_b = np.append(species, species_b)
        offsets_b = np.append(np.zeros((natom, 3)), offsets_b, axis=0)

    # print('build boundary: {:1.2f}'.format(time() - tstart))
    boundary_data = {
        "positions": positions_b,
        "indices": indices_b,
        "species": species_b,
        "offsets": offsets_b,
    }
    return boundary_data



"""
Use ASE's build function
https://wiki.fysik.dtu.dk/ase/ase/build/build.html?highlight=nanotube#ase.build.nanotube
"""

import bpy
from bpy.types import Operator
from bpy.props import (
    StringProperty,
    IntProperty,
    FloatProperty,
    BoolProperty,
)
from ase.build import nanotube
from batoms import Batoms


class BuildNanotube(Operator):
    bl_idname = "nano.nanotube_add"
    bl_label = "Add Nanotube"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add Nanotube"

    label: StringProperty(name="Label", default="tube", description="Label")

    symbol: StringProperty(
        name="Symbol",
        default="C",
        description="The chemical symbol of the element to use.",
    )

    n: IntProperty(name="n", default=6, min=1, soft_max=20, description="System size")

    m: IntProperty(name="m", default=0, min=0, soft_max=20, description="System size")

    length: IntProperty(
        name="Length", default=4, min=1, soft_max=20, description="System size"
    )

    bond: FloatProperty(
        name="bond", default=1.4, min=0, soft_max=15, description="bond"
    )

    vacuum: FloatProperty(
        name="vacuum", default=5.0, min=0, soft_max=15, description="vacuum"
    )

    verbose: BoolProperty(name="verbose", default=True, description="verbose")

    def execute(self, context):
        if self.label == "":
            self.label = self.symbol
        atoms = nanotube(
            n=self.n,
            m=self.m,
            length=self.length,
            bond=self.bond,
            symbol=self.symbol,
            vacuum=self.vacuum,
            verbose=self.verbose,
        )
        batoms = Batoms(label=self.label, from_ase=atoms)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report({"INFO"}, "Add nanotube {}".format(self.label))
        return {"FINISHED"}



import bpy


def test_batoms(ch4):
    """Batoms panel"""
    bpy.ops.object.select_all(action="DESELECT")
    bpy.context.view_layer.objects.active = ch4.obj
    # model_style
    assert bpy.context.scene.batoms.batoms.model_style.upper() == "SPACE-FILLING"
    bpy.context.scene.batoms.batoms.model_style = "Ball-and-stick"
    assert ch4.model_style == 1
    # radius_style
    assert bpy.context.scene.batoms.batoms.radius_style.upper() == "COVALENT"
    bpy.context.scene.batoms.batoms.radius_style = "VDW"
    assert ch4.radius_style == "1"
    # color_style
    assert bpy.context.scene.batoms.batoms.color_style.upper() == "JMOL"
    bpy.context.scene.batoms.batoms.color_style = "CPK"
    assert ch4.color_style == "2"
    # show
    assert bpy.context.scene.batoms.batoms.show is True
    bpy.context.scene.batoms.batoms.show = False
    assert ch4.show[0] == 0



from batoms.base.object import childObjectGN
import numpy as np
import logging

logger = logging.getLogger(__name__)


class SliceBatoms(childObjectGN):
    def __init__(self, label, indices=None, batoms=None) -> None:
        """SliceBatom class

        Two cases:
        1) len(indices) > 1: return Batoms[indices]
        1) one atoms: access properties of one atoms by reading attribute
        from obj directly. This will be faster than Batoms[indices]

        Args:
            label (str):
                Name of the Batoms object
            indices (int, optional):
                indices of the atom. Defaults to 0.
            batoms (Batoms, optional):
                Batoms object. Defaults to None.
        """
        childObjectGN.__init__(self, label, indices, parent=batoms)

    def __repr__(self):
        # bpy.context.view_layer.objects.active = self.obj
        # mode = self.obj.mode
        # bpy.ops.object.mode_set(mode='OBJECT')
        s = "SliceBatoms(species = %s, " % np.unique(self.species)
        # s += "elements = %s, " % str(self.elements)
        s += "positions = %s)" % np.round(self.position, 2)
        # bpy.ops.object.mode_set(mode=mode)
        return s

    @property
    def species(self):
        return self.get_attribute("species")

    # @species.setter
    # def species(self, species):
    #     self.set_species(species)

    # def set_species(self, species):
    #     # self.attributes['species'].data[self.indices].value = species
    #     self.parent.replace([self.indices], species)

    # @property
    # def elements(self):
    #     elements = self.parent.species[self.species].elements
    #     occupancies = {}
    #     for name, eledata in elements.items():
    #         occupancies[name] = round(eledata["occupancy"], 3)
    #     return occupancies

    @property
    def polyhedra(self):
        model_style = self.get_attribute("model_style")
        return model_style == 2

    @polyhedra.setter
    def polyhedra(self, polyhedra):
        if len(self.indices) == 1:
            self.attributes["model_style"].data[self.indices[0]].value = (
                2 if polyhedra else 1
            )
        else:
            self.parent.set_attribute_with_indices(
                "model_style", self.indices, [2 if polyhedra else 1] * len(self.indices)
            )
        self.parent.draw_polyhedra()

    @property
    def bond(self):
        model_style = self.get_attribute("model_style")
        return model_style > 0

    @bond.setter
    def bond(self, bond):
        model_style = np.array(bond).astype(int)
        self.set_attribute("model_style", model_style)
        self.parent.draw_ball_and_stick()



import bpy
import numpy as np
from batoms.base.collection import Setting
from batoms.base.object import BaseObject


class Species(BaseObject):
    """_summary_

    Args:
        BaseObject (_type_): _description_
    """

    def __init__(
        self,
        name,
        parent,
        data=None,
    ) -> None:
        self.name = name
        self.parent = parent
        self.label = parent.label
        self.coll_name = "%s_instancer" % (self.label)
        BaseObject.__init__(
            self, obj_name="%s_%s" % (self.label, name), btype="binstancer"
        )
        if data is not None:
            self.update(data)

    def update(self, data):
        """Update species by new data

        Args:
            data (dict): properites of species
        """
        data["elements"] = self.check_occupancy(data["elements"])
        #
        sp = self.data
        for key, value in data.items():
            if key == "elements":
                continue
            setattr(sp, key, value)
        # add elements
        sp.elements.clear()
        for name, eledata in data["elements"].items():
            ele = sp.elements.add()
            ele.name = name
            if "occupancy" in eledata:
                ele.occupancy = eledata["occupancy"]
            if "color" in eledata:
                ele.color = eledata["color"]
        # update the color for species, which is the color of the main element
        for name, eledata in data["elements"].items():
            if name == self.main_element and "color" in eledata:
                sp.color = eledata["color"]
        self.build_instancer()

    @staticmethod
    def check_occupancy(elements):
        """Check input elements

        Args:
            elements (str, dict): information of elements

        Raises:
            ValueError: Total occupancy should be 1

        Returns:
            _type_: dict of element with its occupancy
        """
        if isinstance(elements, str):
            elements = {elements: {"occupancy": 1.0, "color": (0, 0.2, 0.8, 1)}}
        elif isinstance(elements, dict):
            for ele, eledata in elements.items():
                if isinstance(eledata, (int, float)):
                    elements[ele] = {"occupancy": eledata}
            occupancy = [ele["occupancy"] for ele in elements.values()]
            total = sum(occupancy)
            # not fully occupied.
            if total < 1 - 1e-6:
                elements["X"] = {"occupancy": 1 - total, "color": (0.8, 0.8, 0.0, 1.0)}
            elif total > 1 + 1e-6:
                raise ValueError("Total occupancy should be smaller than 1!")
        return elements

    def build_instancer(self):
        """Build object instancer for species

        Returns:
            bpy Object: instancer
        """
        sp = self.data
        name = "%s_%s" % (self.label, sp.name)
        radius = sp.radius * sp.scale
        segments = sp.segments
        shape = "UV_SPHERE"
        self.delete_obj(name)
        if shape.upper() == "UV_SPHERE":
            bpy.ops.mesh.primitive_uv_sphere_add(
                segments=segments[0], ring_count=segments[1], radius=radius
            )
        if shape.upper() == "METABALL":
            bpy.ops.object.metaball_add(type="BALL", location=[0, 0, 0])
        obj = bpy.context.view_layer.objects.active
        # In the geometry node, a new scale will be add to the instance
        # Here we use a small value to hide the instance from preview
        obj.scale = [0.001, 0.001, 0.001]
        obj.name = name
        obj.data.name = name
        obj.batoms.atom.radius = radius
        obj.batoms.type = "INSTANCER"
        #
        obj.users_collection[0].objects.unlink(obj)
        bpy.data.collections["%s_instancer" % self.label].objects.link(obj)
        bpy.ops.object.shade_smooth()
        if shape.upper() != "METABALL":
            obj.hide_set(True)
            obj.hide_render = True
        #
        self.build_materials(material_style=sp.material_style)
        self.assign_materials()
        # self.color = sp.color
        bpy.context.view_layer.update()
        self.parent.batoms.add_species_node(sp.name, obj)
        # update boundary and search_bond
        if hasattr(self.parent.batoms, "_species"):
            # ignore for when init species
            # only for update
            self.parent.batoms.boundary.update_geometry_node_instancer(sp.name, obj)
            self.parent.batoms.bond.search_bond.update_geometry_node_instancer(
                sp.name, obj
            )
        return obj

    def build_materials(self, node_inputs=None, material_style="default"):
        """Create materials for instancer object

        Args:
            node_inputs (dict, optional):
                properties of materials. Defaults to None.
            material_style (str, optional):
                Materials style. Defaults to 'default'.
        """
        from batoms.material import create_material

        mesh = self.obj.data
        mesh.materials.clear()
        elements = self.elements
        i = 0
        for occ in self.sorted_occupancies:
            ele = elements[occ[0]]
            name = "%s_%s_%s" % (self.label, self.name, ele["name"])
            self.delete_material(name)
            if i == 0:
                color = self.data.color
            else:
                color = ele["color"]
            mat = create_material(
                name,
                color=color,
                node_inputs=node_inputs,
                material_style=material_style,
                backface_culling=True,
            )
            # add attribute and color ramp
            Attrribute = mat.node_tree.nodes.new("ShaderNodeAttribute")
            Attrribute.attribute_type = "INSTANCER"
            ValToRGB = mat.node_tree.nodes.new("ShaderNodeValToRGB")
            ValToRGB.color_ramp.elements.new(0.5)
            ValToRGB.color_ramp.elements[0].color = (1, 0, 0, 1)
            ValToRGB.color_ramp.elements[1].color = (0, 1, 0, 1)
            ValToRGB.color_ramp.elements[2].color = (0, 0, 1, 1)
            mat.node_tree.links.new(Attrribute.outputs["Fac"], ValToRGB.inputs["Fac"])

            mesh.materials.append(mat)
            self.parent.batoms.obj.data.materials.append(mat)
            i += 1

    def assign_materials(self):
        """Assign materials for instancer with order"""
        # find the face index for ele
        mesh = self.obj.data
        occs = self.sorted_occupancies
        nele = len(occs)
        # for occupancy
        if nele > 1:
            # calc the angles
            npoly = len(mesh.polygons)
            normals = np.zeros(npoly * 3)
            material_indexs = np.zeros(npoly, dtype="int")
            mesh.polygons.foreach_get("normal", normals)
            mesh.polygons.foreach_get("material_index", material_indexs)
            normals = normals.reshape(-1, 3)
            xy = normals - np.dot(normals, [0, 0, 1])[:, None] * [0, 0, 1]
            angles = (np.arctan2(xy[:, 1], xy[:, 0]) + np.pi) / np.pi / 2
            #
            tos = 0
            for i in range(1, nele):
                toe = tos + occs[i][1]
                index = np.where((angles > tos) & (angles < toe))[0]
                material_indexs[index] = i
                tos = toe
            mesh.polygons.foreach_set("material_index", material_indexs)

    @property
    def data(self):
        """Get data for a this species

        Returns:
            Bspecies: Class Bspecies
        """
        data = self.parent.find(self.name)
        return data

    @property
    def occupancies(self):
        """Get occupancies for all elements

        Returns:
            dict: Occupancies for all elements for this species
        """
        occupancies = {}
        data = self.data.elements
        for eledata in data:
            occupancies[eledata.name] = round(eledata.occupancy, 3)
        return occupancies

    @property
    def sorted_occupancies(self):
        """Sort occupancies

        Returns:
            tuple: Sorted occupancies
        """
        # sort
        occupancies = self.occupancies
        sorted_occupancies = sorted(occupancies.items(), key=lambda x: -x[1])
        return sorted_occupancies

    @property
    def elements(self):
        """Elements of this species

        Returns:
            dict: dict of Belement
        """
        elements = {}
        for ele in self.data.elements:
            elements[ele.name] = ele.as_dict()
        return elements

    @elements.setter
    def elements(self, elements):
        """Set elements

        Args:
            elements (dict): elements with occupancis

        >>> H2O['O'].elements = 'O'
        >>> H2O['O'].elements = {'O':1.0}
        >>> H2O['O'].elements = {'O':0.7, 'S': 0.3}
        >>> H2O["O"].elements = {"O":{"occupancy": 0.7},
                            "S":{"occupancy"}}
        """
        data = self.data.as_dict()
        if isinstance(elements, str):
            elements = {"elements": {elements: {"occupancy": 1.0}}}
        if isinstance(elements, dict):
            for ele, eledata in elements.items():
                if isinstance(eledata, (int, float)):
                    elements[ele] = {"occupancy": eledata}
        data["elements"] = elements
        self.update(data)

    @property
    def main_element(self):
        """Main element

        Returns:
            str: name of the element with largest occupancies,
            but vacancy "X" is not included.
        """
        main_element = {}
        occs = self.sorted_occupancies
        if occs[0][0] == "X" and len(occs) > 1:
            main_element = occs[1][0]
        else:
            main_element = occs[0][0]
        return main_element

    @property
    def materials(self):
        """Get materials

        Returns:
            dict: dict of bpy material object
        """
        materials = {}
        for ele in self.elements:
            name = "%s_%s_%s" % (self.label, self.name, ele)
            mat = bpy.data.materials.get(name)
            materials[ele] = mat
        return materials

    @materials.setter
    def materials(self, materials):
        """Set materials properties of the instancer

        Args:
            materials (dict): parameters for materials

        >>> h2o['H'].materials = {'Metallic': 0.9,
        ...     'Specular': 1.0, 'Roughness': 0.01}
        """
        self.build_materials(node_inputs=materials)
        self.assign_materials()

    @property
    def color(self):
        return self.get_color()

    @color.setter
    def color(self, color):
        """
        >>> h.color = [0.8, 0.1, 0.3, 1.0]
        """
        self.set_color(color)

    def get_color(self):
        """ """
        # Viewpoint_color = self.materials[self.main_element].diffuse_color
        for node in self.materials[self.main_element].node_tree.nodes:
            if "Base Color" in node.inputs:
                node_color = node.inputs["Base Color"].default_value[:]
            if "Alpha" in node.inputs:
                Alpha = node.inputs["Alpha"].default_value
        color = [node_color[0], node_color[1], node_color[2], Alpha]
        return color

    def set_color(self, color):
        if len(color) == 3:
            color = [color[0], color[1], color[2], 1]
        self.materials[self.main_element].diffuse_color = color
        for node in self.materials[self.main_element].node_tree.nodes:
            if "Base Color" in node.inputs:
                node.inputs["Base Color"].default_value = color
            if "Alpha" in node.inputs:
                node.inputs["Alpha"].default_value = color[3]
        # self.data.elements[self.main_element].color = color
        self.data.color = color

    @property
    def material_style(self):
        return self.data.material_style

    @material_style.setter
    def material_style(self, material_style):
        self.data.material_style = material_style
        self.build_materials(material_style=material_style)
        self.assign_materials()

    def color_by_attribute(self, attribute, cmap="viridis"):
        """Color by attribute"""
        import matplotlib.pyplot as plt

        node_tree = self.materials[self.main_element].node_tree
        if attribute in ["element"]:
            # remove the link
            for link in node_tree.links:
                if (
                    link.from_node.name in ["Color Ramp", "ColorRamp"]
                    and link.to_node.name == "Principled BSDF"  # noqa: W503
                ):
                    node_tree.links.remove(link)
        else:
            # Blender 3.6: 'Color Ramp', Blender 3.4: 'ColorRamp'
            if node_tree.nodes.get("Color Ramp"):
                color_ramp_node = node_tree.nodes.get("Color Ramp")
            else:
                color_ramp_node = node_tree.nodes.get("ColorRamp")
            node_tree.nodes["Attribute"].attribute_name = attribute
            colormap = plt.get_cmap(cmap)
            colors = colormap([0, 0.5, 1])
            color_ramp_node.color_ramp.elements[0].color = colors[0]
            color_ramp_node.color_ramp.elements[1].color = colors[1]
            color_ramp_node.color_ramp.elements[2].color = colors[2]
            node_tree.links.new(
                color_ramp_node.outputs["Color"],
                node_tree.nodes["Principled BSDF"].inputs["Base Color"],
            )

    @property
    def radius(self):
        return self.get_radius()

    def get_radius(self):
        return self.data.radius

    @property
    def indices(self):
        indices = np.where(self.parent.batoms.attributes["species"] == self.name)[0]
        return indices

    @property
    def scale(self):
        return self.parent.batoms.get_attribute_with_indices("scale", self.indices)

    @scale.setter
    def scale(self, scale):
        self.parent.batoms.set_attribute_with_indices("scale", self.indices, scale)

    @property
    def segments(self):
        return self.get_segments()

    @segments.setter
    def segments(self, segments):
        self.set_segments(segments)

    def get_segments(self):
        return self.data.segmetns[:]

    def set_segments(self, segments):
        if isinstance(segments, int):
            segments = [segments, segments]
            # raise Exception('Segments should be int!')
        self.data.segments = segments
        self.update(self.data.as_dict())

    def as_dict(self) -> dict:
        data = {
            "species": self.name,
            "name": self.name,
            "radius": self.radius,
            "elements": self.elements,
        }
        return data

    def __repr__(self):
        s = "Bspecies(species = '%s', " % self.name
        s += "elements = %s, " % str(self.sorted_occupancies)
        s += "color = %s)" % np.round(np.array(self.color), 2)
        # s += "radius_style = %s, " % self.radius_style
        # s += "color_style = %s)" % self.color_style
        return s


class Bspecies(Setting):
    """Bspecies Class"""

    def __init__(
        self,
        label,
        species=None,
        batoms=None,
        material_style="default",
        segments=None,
        subdivisions=2,
    ) -> None:
        Setting.__init__(self, label, coll_name=label)
        self.label = label
        self.name = "batoms"
        self.batoms = batoms
        #
        self.calc_segments(segments)
        self.subdivisions = subdivisions
        if species is not None:
            self.add(species)
        # for sp, data in species.items():
        # self[sp] = data

    def get_ui_list_index(self):
        return self.bpy_data.ui_list_index_species

    def set_ui_list_index(self, value):
        self.bpy_data.ui_list_index_species = value

    def get_bpy_setting(self):
        if self.coll_name:
            coll = bpy.data.collections.get(self.coll_name)
            data = getattr(coll, self.name)
        else:
            raise KeyError("The collection property {} not exist!".format(self.name))
        return data.settings_species

    def calc_segments(self, segments):
        if segments is not None:
            self.segments = segments
            return
        if self.batoms is None:
            self.segments = [32, 24]
            return
        natom = len(self.batoms)
        if natom <= 1e3:
            segments = [32, 24]
        elif natom <= 1e4:
            segments = [16, 16]
        elif natom <= 1e4:
            segments = [10, 10]
        elif natom <= 1e5:
            segments = [8, 8]
        else:
            segments = [6, 6]
        self.segments = segments

    def __getitem__(self, name):
        if isinstance(name, str):
            sp = Species(name, parent=self)
        else:
            raise ValueError("Please input a string")
        return sp

    def __setitem__(self, name, data):
        """
        Set species
        """
        self.add({name: data})

    def add(self, props, instancer=None):
        """ """
        from batoms.utils import get_default_species_data

        if props is None:
            return
        if isinstance(props, str):
            props = [props]
        if isinstance(props, (list, np.ndarray)):
            species_list = props
            props = {}
            for species in species_list:
                ele = species.split("_")[0]
                props[species] = get_default_species_data(
                    {ele: {"occupancy": 1.0}},
                    radius_style=self.batoms.radius_style,
                    color_style=self.batoms.color_style,
                )

        for name, data in props.items():
            # if 'color_style' not in data:
            # data['color_style'] = '0'
            # if 'radius_style' not in data:
            # data['radius_style'] = '0'
            sp = self.find(name)
            if sp is None:
                sp = self.bpy_setting.add()
                self.ui_list_index = len(self) - 1
            if "color" not in data:
                props = get_default_species_data(
                    data["elements"],
                    radius_style=self.batoms.radius_style,
                    color_style=self.batoms.color_style,
                )
                data.update(props)
            sp.name = name
            sp.label = self.label
            sp.segments = self.segments
            sp = Species(sp.name, parent=self, data=data)

    def update_geometry_node(self):
        instancers = self.instancers
        for sp, obj in instancers.items():
            self.batoms.add_species_node(sp, obj)

    def keys(self):
        return self.species.keys()

    def items(self):
        return self.species.items()

    def __repr__(self):
        s = ""
        for name, sp in self.species.items():
            s += str(sp)
            s += "\n"
        return s

    @property
    def instancers(self):
        return self.get_instancers()

    def get_instancers(self):
        instancers = {}
        for sp in self.species:
            name = "%s_%s" % (self.label, sp)
            instancers[sp] = bpy.data.objects.get(name)
        return instancers

    @property
    def species(self):
        return self.get_species()

    def get_species(self):
        species = {}
        for sp in self.bpy_setting:
            species[sp.name] = Species(sp.name, parent=self)
        return species

    @property
    def main_elements(self):
        main_elements = {}
        for name, sp in self.species.items():
            main_elements[name] = sp.main_element
        return main_elements

    @property
    def species_props(self):
        return self.get_species_props()

    def get_species_props(self):
        species_props = {}
        for name, sp in self.species.items():
            radius = sp.radius
            color = sp.color
            species_props[name] = {
                "radius": radius,
                "color": color,
                "element": sp.main_element,
            }
        return species_props

    def __add__(self, other):
        self += other
        return self

    def __iadd__(self, other):
        self.extend(other)
        return self

    def copy(self, label):
        """Copy this Bspecies class

        Args:
            name (str): new label

        Returns:
            Class Bspecies: copy
        """
        for sp, instancer in self.instancers.items():
            instancer = instancer.copy()
            bpy.data.collections["Collection"].objects.link(instancer)
            instancer.hide_set(True)
            instancer.name = "%s_%s_instancer" % (label, sp)
        bspecies = self.__class__(label, label, species={}, batoms=None)
        return bspecies

    def extend(self, other):
        """ """
        species1 = self.species
        species2 = other.species
        species3 = {}
        for key, sp in species2.items():
            data = sp.as_dict()
            if key not in species1:
                instancer = other.instancers[key]
                name = "%s_%s" % (self.label, key)
                instancer.name = name
                # materials
                # for mat in instancer.materials_slots:
                # mat.name.replace()
                species3[key] = [data, instancer]
            elif data != species1[key].as_dict():
                instancer = other.instancers[key]
                spname = "%s_%s" % (key, other.label)
                name = "%s_%s" % (self.label, spname)
                instancer.name = name
                species3[spname] = [data, instancer]
        #
        for key, data in species3.items():
            self.add({key: data[0]}, instancer=data[1])
        return self

    def __iter__(self):
        item = self.bpy_setting
        for i in range(len(item)):
            yield item[i]

    def update(self, segments=None):
        if segments:
            self.segments = segments
        for name, sp in self.species.items():
            sp.build_instancer()

    def as_dict(self) -> dict:
        species = self.species
        data = {}
        for name, sp in species.items():
            data[name] = sp.as_dict()
        return data

    def color_by_attribute(self, attribute, cmap="viridis"):
        for _, sp in self.species.items():
            sp.color_by_attribute(attribute, cmap=cmap)
